/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "CPU.h"
#include "Interrupt.h"
#include "RSP_HLE.h"
#include "RSP.h"
#include "Memory.h"

#include "Debug/Debug.h"
#include "Debug/Dump.h"			// For Dump_GetDumpDirectory()

#include "OSHLE/ultra_rcp.h"
#include "OSHLE/ultra_mbi.h"
#include "OSHLE/ultra_sptask.h"

#include "Debug/DBGConsole.h"
#include "Interface/MainWindow.h"

#include "Plugins/AudioPlugin.h"
#include "Plugins/GraphicsPlugin.h"

#include "Utility/Profiler.h"
#include "Utility/PrintOpCode.h"

#ifdef DAEDALUS_PSP
#include "PSPGraphics/DaedalusGraphics.h"
#endif

#if 0
static void RDP_DumpRSPCode(char * szName, u32 dwCRC, u32 * pBase, u32 dwPCBase, u32 dwLen);
static void RDP_DumpRSPData(char * szName, u32 dwCRC, u32 * pBase, u32 dwPCBase, u32 dwLen);
#endif


//*****************************************************************************
// Defines
//*****************************************************************************
#undef LLE_RSP
#define HLE_AUDIO
#define HLE_GRAPHICS

//*****************************************************************************
// Static Variables
//*****************************************************************************
static OSTask	* g_pTask = NULL;

//*****************************************************************************
//
//*****************************************************************************
void RSP_HLE_EnableGfx()
{
	g_bRDPEnableGfx = TRUE;
}

//*****************************************************************************
//
//*****************************************************************************
void RSP_HLE_DisableGfx()
{
	g_bRDPEnableGfx = FALSE;
}

//*****************************************************************************
// Return true if we have a pending RSP task
//*****************************************************************************
bool RSP_HLE_IsTaskLoaded()
{
	return g_pTask ? true : false;
}

#if 0
//*****************************************************************************
//
//*****************************************************************************
void RDP_DumpRSPCode(char * szName, u32 dwCRC, u32 * pBase, u32 dwPCBase, u32 dwLen)
{
	TCHAR opinfo[400];
	TCHAR szFilePath[MAX_PATH+1];
	TCHAR szFileName[MAX_PATH+1];
	FILE * fp;

	Dump_GetDumpDirectory(szFilePath, "rsp_dumps");

	sprintf(szFileName, "task_dump_%s_crc_0x%08x.txt", szName, dwCRC);

	IO::Path::Append(szFilePath, szFileName);
	

	fp = fopen(szFilePath, "w");
	if (fp == NULL)
		return;

	u32 dwIndex;
	for (dwIndex = 0; dwIndex < dwLen; dwIndex+=4)
	{
		OpCode op;
		u32 pc = dwIndex&0x0FFF;
		op._u32 = pBase[dwIndex/4];

		SprintRSPOpCodeInfo( opinfo, pc + dwPCBase, op );

		fprintf(fp, "0x%08x: <0x%08x> %s\n", pc + dwPCBase, op._u32, opinfo);
		//fprintf(fp, "<0x%08x>\n", dwOpCode);
	}
	
	fclose(fp);
}
#endif

#if 0
//*****************************************************************************
//
//*****************************************************************************
void RDP_DumpRSPData(char * szName, u32 dwCRC, u32 * pBase, u32 dwPCBase, u32 dwLen)
{
	TCHAR szFilePath[MAX_PATH+1];
	TCHAR szFileName[MAX_PATH+1];
	FILE * fp;

	Dump_GetDumpDirectory(szFilePath, "rsp_dumps");

	sprintf(szFileName, "task_data_dump_%s_crc_0x%08x.txt", szName, dwCRC);

	IO::Path::Append(szFilePath, szFileName);

	fp = fopen(szFilePath, "w");
	if (fp == NULL)
		return;

	u32 dwIndex;
	for (dwIndex = 0; dwIndex < dwLen; dwIndex+=4)
	{
		u32 dwData;
		u32 pc = dwIndex&0x0FFF;
		dwData = pBase[dwIndex/4];

		fprintf(fp, "0x%08x: 0x%08x\n", pc + dwPCBase, dwData);
	}
	
	fclose(fp);

}
#endif

//*****************************************************************************
//
//*****************************************************************************
#if 0
static void	RSP_HLE_DumpTaskInfo( const OSTask * pTask )
{
	DBGConsole_Msg(0, "DP: Task:%08x Flags:%08x BootCode:%08x BootCodeSize:%08x",
		pTask->t.type, pTask->t.flags, pTask->t.ucode_boot, pTask->t.ucode_boot_size);

	DBGConsole_Msg(0, "DP: uCode:%08x uCodeSize:%08x uCodeData:%08x uCodeDataSize:%08x",
		pTask->t.ucode, pTask->t.ucode_size, pTask->t.ucode_data, pTask->t.ucode_data_size);

	DBGConsole_Msg(0, "DP: Stack:%08x StackS:%08x Output:%08x OutputS:%08x",
		pTask->t.dram_stack, pTask->t.dram_stack_size, pTask->t.output_buff, pTask->t.output_buff_size);

	DBGConsole_Msg(0, "DP: Data(PC):%08x DataSize:%08x YieldData:%08x YieldDataSize:%08x",
		pTask->t.data_ptr, pTask->t.data_size, pTask->t.yield_data_ptr, pTask->t.yield_data_size);
}
#endif


//*****************************************************************************
//
//*****************************************************************************
void RSP_HLE_CheckForTaskLoad()
{
	OSTask * pTask = (OSTask *)(g_pu8SpMemBase + 0x0FC0);

#ifdef LLE_RSP

	extern bool gStopOnNextRSP;

	switch (pTask->t.type)
	{
		case M_GFXTASK:		DBGConsole_Msg( 0, "RSP: [WGraphics]" ); /*gStopOnNextRSP = true;*/ 
			
			//RSP_HLE_DumpTaskInfo( pTask );
			
			break;
		case M_AUDTASK:		/*DBGConsole_Msg( 0, "RSP: [WAudio]" );*/	break;
		case M_VIDTASK:		DBGConsole_Msg( 0, "RSP: [WVideo]" ); break;
		case M_JPGTASK:		DBGConsole_Msg( 0, "RSP: [WJPEG]" ); break;
	}

#else

	bool handle_task = false;

	switch ( pTask->t.type )
	{
		case M_GFXTASK:			// Handle only if HLE_GRAPHICS is defined
								#ifdef HLE_GRAPHICS
								handle_task = true;
								#endif
								break;

		case M_AUDTASK:			// Handle only if HLE_AUDIO is defined
								#ifdef HLE_AUDIO
								handle_task = true;
								#endif
								break;

		case M_VIDTASK:			// Can't handle
								break;

		case M_JPGTASK:			// Can't handle
								break;

		default:				// Can't handle	
								DBGConsole_Msg(0, "Unknown task: %08x", pTask->t.type );
							//	RSP_HLE_DumpTaskInfo( pTask );
							//	RDP_DumpRSPCode("boot",    0xDEAFF00D, (u32*)(g_pu8RamBase + (((u32)pTask->t.ucode_boot)&0x00FFFFFF)), 0x04001000, pTask->t.ucode_boot_size);
							//	RDP_DumpRSPCode("unkcode", 0xDEAFF00D, (u32*)(g_pu8RamBase + (((u32)pTask->t.ucode)&0x00FFFFFF)),      0x04001080, 0x1000 - 0x80);//pTask->t.ucode_size);
								break;
	}

	//
	// If we want to handle the task, set the pointer
	// 
	if ( handle_task )
	{
		g_pTask = pTask;
	}
	else
	{
		g_pTask = NULL;
	}

#endif		// LLE_RSP
}




//*****************************************************************************
//
//*****************************************************************************
void RSP_HLE_ExecuteTask()
{
#ifdef LLE_RSP

	DAEDALUS_ASSERTMSG( "We should not be in RSP_HLE_ExecuteTask() when LLE_RSP is defined" );

#else
	DAEDALUS_ASSERT( g_pTask, "Called RSP_HLE_ExecuteTask with no task ready." );

	switch ( g_pTask->t.type )	
	{		
		case M_GFXTASK:
							#ifdef HLE_GRAPHICS

					#ifdef DAEDALUS_PSP
							//printf( "Dlist\n" );
							PSPGraphics_ProcessDisplayList();
					#else
								if ( CGraphicsPlugin::IsAvailable() && g_bRDPEnableGfx )
								{
								#ifdef DAEDALUS_USE_EXCEPTIONS
									try
								#endif
									{
										DAEDALUS_PROFILE( "HLE: Graphics" );
										CGraphicsPlugin::Get()->ProcessDList();
									}
								#ifdef DAEDALUS_USE_EXCEPTIONS
									catch (...)
									{
										if (g_DaedalusConfig.TrapExceptions)
										{
											DBGConsole_Msg(0, "Exception in GFX processing");
											Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_DP);
											R4300_Interrupt_UpdateCause3();
										}
										else
										{
											// Give us the option of continuing
											u32 result( CMainWindow::Get()->MessageBox("Exception in GFX processing - Throw Exception?", g_szDaedalusName, MB_ICONEXCLAMATION|MB_YESNO) );
											if (result == IDYES)
											{
												throw;
											}
											else
											{
												Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_DP);
												R4300_Interrupt_UpdateCause3();
											}
										}
									}
								#endif
								}
						#endif
							#else
								DAEDALUS_ASSERTMSG( "We should not be trying to execute a HLE Graphics task when HLE_GRAPHICS is not defined" );
							#endif
							break;

		case M_AUDTASK:
							#ifdef HLE_AUDIO
								if (g_pAiPlugin != NULL)
								{
									//RDP_AUD_ExecuteTask(pTask);
									DAEDALUS_PROFILE( "HLE: Audio" );							
									g_pAiPlugin->ProcessAList();
								}
							#else
								DAEDALUS_ASSERTMSG( "We should not be trying to execute a HLE Audio task when HLE_AUDIO is not defined" );
							#endif
							break;
	}

	// Need to point to last instr?
	//Memory_DPC_SetRegister(DPC_CURRENT_REG, (u32)pTask->t.data_ptr);

	g_pTask = NULL;

	//
	// Set the SP flags appropriately. The RSP is not running anyway, no need to stop it
	//
	Memory_SP_SetRegisterBits(SP_STATUS_REG, SP_STATUS_TASKDONE|SP_STATUS_BROKE|SP_STATUS_HALT);

	//
	// We've set the SP_STATUS_BROKE flag - better check if it causes an interrupt
	//
	if( Memory_SP_GetRegister(SP_STATUS_REG) & SP_STATUS_INTR_BREAK )
	{
		Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_SP);
		R4300_Interrupt_UpdateCause3();
	}

#endif		// LLE_RSP
}

