/*
Copyright (C) 2001,2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


#include "stdafx.h"
#include "CodeGeneratorX86.h"
#include "DynaRecInternal.h"
#include "DynaRecFP.h"
#include "Trace.h"
#include "IndirectExitMap.h"

#include "Debug/Debug.h"
#include "Debug/DBGConsole.h"

#include "Core/CPU.h"
#include "Core/Registers.h"

#include "OSHLE/ultra_R4300.h"

#include "AssemblyUtils.h"

// XXXX
#include "DR_FP.inl"

using namespace AssemblyUtils;

static const u32		NUM_MIPS_REGISTERS( 32 );
static const u32		INTEL_REG_UNUSED( ~0 );



//*****************************************************************************
//	XXXX
//*****************************************************************************
void Dynarec_ClearedCPUStuffToDo()
{
}
void Dynarec_SetCPUStuffToDo()
{
}

//*****************************************************************************
//	Register Caching
//*****************************************************************************
class CRegisterStatus
{
	struct CachedRegInfo
	{
		EIntelReg iCachedIReg;	// INVALID_CODE if uncached, else ???_CODE of intel reg we're cached in
		bool	BLoValid;		// If cached, TRUE if value in intel register is valid
		bool	BLoDirty;		// If cached, TRUE if value has been modified

		u32		HiValue;		// If BHiUnknown is FALSE, this is the value of the register (through setmipshi..)
		bool	BHiDirty;		// Do we need to write information about this register back to memory?
		bool	BHiUnknown;		// If TRUE, we don't know the value of this register
	};
public:

	void Reset()
	{
		for (u32 i = 0; i < NUM_REGISTERS; i++)
		{
			SetCachedReg( i, INVALID_CODE );
			MarkAsValid( i, false );
			MarkAsDirty( i, false );

			MarkHiAsDirty( i, false );
			MarkHiAsUnknown( i, true );
			SetHiValue( i, 0 );		// Ignored
		}
	}

	inline void MarkAsValid( u32 mreg, bool valid )
	{
		mCachedRegisterInfo[mreg].BLoValid = valid;
	}

	inline bool IsValid( u32 mreg ) const
	{
		return mCachedRegisterInfo[mreg].BLoValid;
	}

	inline bool IsDirty( u32 mreg ) const
	{
		return mCachedRegisterInfo[mreg].BLoDirty;
	}

	inline void MarkAsDirty( u32 mreg, bool dirty )
	{
		mCachedRegisterInfo[mreg].BLoDirty = dirty;	
	}

	inline EIntelReg GetCachedReg( u32 mreg ) const
	{
		return mCachedRegisterInfo[mreg].iCachedIReg;
	}

	inline void SetCachedReg( u32 mreg, EIntelReg reg )
	{
		mCachedRegisterInfo[mreg].iCachedIReg = reg;
	}

	inline void MarkHiAsUnknown( u32 mreg, bool unk )
	{
		mCachedRegisterInfo[mreg].BHiUnknown = unk;
	}

	inline void MarkHiAsDirty( u32 mreg, bool dirty )
	{
		mCachedRegisterInfo[mreg].BHiDirty = dirty;	
	}

	inline bool IsHiDirty( u32 mreg ) const
	{
		return mCachedRegisterInfo[mreg].BHiDirty;
	}
	
	inline bool IsHiUnknown( u32 mreg ) const
	{
		return mCachedRegisterInfo[mreg].BHiUnknown;
	}

	inline void SetHiValue( u32 mreg, u32 value )
	{
		mCachedRegisterInfo[mreg].HiValue = value;
	}

	inline u32 GetHiValue( u32 mreg ) const
	{
		return mCachedRegisterInfo[mreg].HiValue;
	}

private:
	static const u32	NUM_REGISTERS = 32;
	CachedRegInfo		mCachedRegisterInfo[NUM_REGISTERS];
};


static CRegisterStatus	gRegisterStatus;
static u32				gIntelRegUsageMap[NUM_X86_REGISTERS];
static u32				gWriteCheck[NUM_MIPS_REGISTERS];

//*****************************************************************************
//
//*****************************************************************************
CCodeGeneratorX86::CCodeGeneratorX86( CAssemblyBuffer * p_primary, CAssemblyBuffer * p_secondary )
:	CCodeGenerator( )
,	CAssemblyWriterX86( p_primary )
,	mSpCachedInESI( false )
,	mSetSpPostUpdate( 0 )
,	mpPrimary( p_primary )
,	mpSecondary( p_secondary )
{
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::Finalise()
{
	SetAssemblyBuffer( NULL );
	mpPrimary = NULL;
	mpSecondary = NULL;
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::AnalyseStats( const SRegisterStats & stats )
{
	// Determine which base reg to cache
	u32 dwMaxBaseUseCount = 0;
	u32 dwMaxBaseUseIndex = INTEL_REG_UNUSED;
	// Determine which general reg to cache
	u32 dwMaxUseCount[3] = { 0,0,0 };
	u32 dwMaxUseIndex[3] = { INTEL_REG_UNUSED,INTEL_REG_UNUSED,INTEL_REG_UNUSED };

	// Ignore reg0 - start at reg 1
	for (u32 i = 1; i < NUM_MIPS_REGISTERS; i++)
	{
		// Base reg suitability?
		if (stats.RegBaseUse[i] > 1 &&
			stats.RegUpdates[i] == 0)
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  Mmm - %s looks nice for base caching with %d uses", 
				RegNames[i], stats.RegBaseUse[i] ) );

			if (stats.RegBaseUse[i] > dwMaxBaseUseCount)
			{
				dwMaxBaseUseCount = stats.RegBaseUse[i];
				dwMaxBaseUseIndex = i;
			}
		}

		// Read reg suitability?
		u32 dwUses = stats.RegReads[i] +
					 stats.RegUpdates[i] + 
					 stats.RegBaseUse[i];

		// TODO: When more load/store ops are optimised, add base usage to count.
		// TODO: Writes are more expensive with more generic ops, becase we have to
		//       re-load the register is its contents have changed in the call!
		if (dwUses > 1)			
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  Mmm - %s looks nice for general caching with %d reads, %d writes, %d base uses", 
				RegNames[i], stats.RegReads[i], stats.RegUpdates[i], stats.RegBaseUse[i] ) );

			if (dwUses > dwMaxUseCount[0])
			{
				dwMaxUseCount[2] = dwMaxUseCount[1];
				dwMaxUseIndex[2] = dwMaxUseIndex[1];
				
				dwMaxUseCount[1] = dwMaxUseCount[0];
				dwMaxUseIndex[1] = dwMaxUseIndex[0];
				
				dwMaxUseCount[0] = dwUses;
				dwMaxUseIndex[0] = i;
			}
			else if (dwUses > dwMaxUseCount[1])
			{
				dwMaxUseCount[2] = dwMaxUseCount[1];
				dwMaxUseIndex[2] = dwMaxUseIndex[1];
				
				dwMaxUseCount[1] = dwUses;
				dwMaxUseIndex[1] = i;
			}
			else if (dwUses > dwMaxUseCount[2])
			{			
				dwMaxUseCount[2] = dwUses;
				dwMaxUseIndex[2] = i;
			}
		}	
	}


	if (dwMaxBaseUseIndex != INTEL_REG_UNUSED)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  Best register for base pointer caching is %s", RegNames[dwMaxBaseUseIndex] ) );
	}
	// Only bother if the register is used as a base more than once
	if (stats.RegBaseUse[REG_sp] > 1)
	{
		// Double the "preference" for base caching of sp, because it avoids
		// having to call ReadAddress
		if (dwMaxUseIndex[2] != INTEL_REG_UNUSED && (stats.RegBaseUse[REG_sp]*2) < dwMaxUseCount[2])
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  Caching %s rather than sp/base",
				RegNames[dwMaxUseIndex[2]], stats.RegUpdates[REG_sp] ) );
		}
		else
		{
			// If the register changes several times, ignore
			if (stats.RegUpdates[REG_sp] > 1)
			{
				DPF( DEBUG_DYNREC, DSPrintf( "  Unable to cache SP - %d updates", stats.RegUpdates[REG_sp] ) );
			}
			else
			{
				DPF( DEBUG_DYNREC, DSPrintf( "  Will cache sp in ESI after update %d", stats.RegUpdates[REG_sp] ) );

				mSetSpPostUpdate = stats.RegUpdates[REG_sp];
			}
		}
	}

	if (dwMaxUseIndex[0] != INTEL_REG_UNUSED)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  Best register for general caching is %s", RegNames[dwMaxUseIndex[0]] ) );

		// Cache read reg in EDI:
		// Save edi for future use
		PUSH(EDI_CODE);

		gRegisterStatus.SetCachedReg( dwMaxUseIndex[0], EDI_CODE );
		gRegisterStatus.MarkAsDirty( dwMaxUseIndex[0], false );
		gRegisterStatus.MarkAsValid( dwMaxUseIndex[0], false );

		gIntelRegUsageMap[EDI_CODE] = dwMaxUseIndex[0];
	}

	// Do this here, as EBX < ESI < EDI 
	if (mSetSpPostUpdate == u32(~0) && dwMaxUseIndex[2] != INTEL_REG_UNUSED)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  3rd best register for general caching is %s", RegNames[dwMaxUseIndex[2]] ) );

		// Cache read reg in ESI:
		// Save esi for future use
		PUSH(ESI_CODE);

		gRegisterStatus.SetCachedReg( dwMaxUseIndex[2], ESI_CODE );
		gRegisterStatus.MarkAsDirty( dwMaxUseIndex[2], false );
		gRegisterStatus.MarkAsValid( dwMaxUseIndex[2], false );
		gIntelRegUsageMap[ESI_CODE] = dwMaxUseIndex[2];
	}

	if (dwMaxUseIndex[1] != INTEL_REG_UNUSED)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  2nd best register for general caching is %s", RegNames[dwMaxUseIndex[1]] ) );

		// Cache read reg in EBX:
		// Save ebx for future use
		PUSH(EBX_CODE);
	
		gRegisterStatus.SetCachedReg( dwMaxUseIndex[1], EBX_CODE );
		gRegisterStatus.MarkAsDirty( dwMaxUseIndex[1], false );
		gRegisterStatus.MarkAsValid( dwMaxUseIndex[1], false );
		gIntelRegUsageMap[EBX_CODE] = dwMaxUseIndex[1];
	}

	if (mSetSpPostUpdate != u32(~0))
		PUSH(ESI_CODE);
}


//*****************************************************************************
//
//*****************************************************************************
#if 0
u32 gNumFragmentsExecuted = 0;
extern "C"
{

void LogFragmentEntry( u32 entry_address )
{
	gNumFragmentsExecuted++;
	if(gNumFragmentsExecuted >= 0x99990)
	{
		OutputDebugString( DSPrintf( "Address %08x\n", entry_address ) );
	}
}

}
#endif

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::Initialise( u32 entry_address, u32 * hit_counter )
{
	//MOVI(ECX_CODE, entry_address);
	// CALL( CCodeLabel( LogFragmentEntry ) );

	if( hit_counter != NULL )
	{
		MOV_REG_MEM( EAX_CODE, hit_counter );
		ADDI( EAX_CODE, 1 );
		MOV_MEM_REG( hit_counter, EAX_CODE );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::SetCachedBasePointer( const void * p_base )
{
	// This is ignored for now
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::SetRegisterSpanList( const RegisterSpanList & span_list )
{
	// This is ignored for now
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::UpdateRegisterCaching( u32 instruction_idx )
{
	// This is ignored for now
}

//*****************************************************************************
//
//*****************************************************************************
RegisterSnapshotHandle	CCodeGeneratorX86::GetRegisterSnapshot()
{
	// This doesn't do anything useful yet.
	return RegisterSnapshotHandle( 0 );
}

//*****************************************************************************
//
//*****************************************************************************
CCodeLabel	CCodeGeneratorX86::GetEntryPoint() const
{
	return mpPrimary->GetStartAddress();
}

//*****************************************************************************
//
//*****************************************************************************
u32	CCodeGeneratorX86::GetCompiledCodeSize() const
{
	return mpPrimary->GetSize() + mpSecondary->GetSize();
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::GeneratePrologue()
{
	// Set the sp when the update count equals this value
	mSpCachedInESI = false;
	mSetSpPostUpdate = u32(~0);

	// Clear the fp status
	SR_FP_Init(this);

	gRegisterStatus.Reset();

	// Init intel regs
	for (u32 i = 0; i < NUM_X86_REGISTERS; i++)
	{
		gIntelRegUsageMap[i] = INTEL_REG_UNUSED;
	}
}

//*****************************************************************************
// EPILOGUE - set up g_pPCMemBase before returning
//*****************************************************************************
CCodeLabel CCodeGeneratorX86::GenerateEpilogue( u32 pc )
{
	CCodeLabel epilogue_label( GetAssemblyBuffer()->GetLabel() );

	MOVI(EAX_CODE, pc);
	MOV_MEM_REG(&gCPUState.CurrentPC, EAX_CODE);
/*
	GenerateReadWriteAddress(pc, 0, 0, false, ~0, true, true);
/*/
	MOV(ECX_CODE, EAX_CODE);
	SHRI(EAX_CODE, 18);

	// call dword ptr [g_ReadAddressLookupTable + eax*4]
	CALL_MEM_PLUS_REGx4( g_ReadAddressLookupTable, EAX_CODE );

/**/
	MOV_MEM_REG(&g_pPCMemBase, EAX_CODE);
	RET();

	return epilogue_label;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation CCodeGeneratorX86::GenerateExitCode( u32 exit_address, u32 jump_address, u32 num_instructions, CCodeLabel next_fragment )
{
	//DAEDALUS_ASSERT( exit_address != u32( ~0 ), "Invalid exit address" );
	DAEDALUS_ASSERT( !next_fragment.IsSet() || jump_address == 0, "Shouldn't be specifying a jump address if we have a next fragment?" );

#ifdef _DEBUG
	if(exit_address == u32(~0))
	{
		INT3();
	}
#endif

	MOVI(ECX_CODE, num_instructions);
	CALL( CCodeLabel( CPU_UpdateCounter ) );

	// This jump may be NULL, in which case we patch it below
	// This gets patched with a jump to the next fragment if the target is later found
	CJumpLocation jump_to_next_fragment( GenerateBranchIfNotSet( &gCPUState.StuffToDo, next_fragment ) );

	// If the flag was set, we need in initialise the pc/delay to exit with
	CCodeLabel interpret_next_fragment( GetAssemblyBuffer()->GetLabel() );

	u8		exit_delay;

	if( jump_address != 0 )
	{
		SetVar( &gCPUState.TargetPC, jump_address );
		exit_delay = EXEC_DELAY;
	}
	else
	{
		exit_delay = NO_DELAY;
	}

	SetVar8( &gCPUState.Delay, exit_delay );
	SetVar( &gCPUState.CurrentPC, exit_address );
	
	// No need to call CPU_SetPC(), as this is handled by CFragment when we exit
	RET();

	// Patch up the exit jump
	if( !next_fragment.IsSet() )
	{
		PatchJumpLong( jump_to_next_fragment, interpret_next_fragment );
	}

	return jump_to_next_fragment;
}

//*****************************************************************************
// Handle branching back to the interpreter after an ERET
//*****************************************************************************
void CCodeGeneratorX86::GenerateEretExitCode( u32 num_instructions, CIndirectExitMap * p_map )
{
	MOVI(ECX_CODE, num_instructions);
	CALL( CCodeLabel( CPU_UpdateCounter ) );

	// We always exit to the interpreter, regardless of the state of gCPUState.StuffToDo

	// Eret is a bit bodged so we exit at PC + 4
	MOV_REG_MEM( EAX_CODE, &gCPUState.CurrentPC );
	ADDI( EAX_CODE, 4 );
	MOV_MEM_REG( &gCPUState.CurrentPC, EAX_CODE );
	SetVar8( &gCPUState.Delay, NO_DELAY );
	
	// No need to call CPU_SetPC(), as this is handled by CFragment when we exit

	RET();
}

//*****************************************************************************
// Handle branching back to the interpreter after an indirect jump
//*****************************************************************************
void CCodeGeneratorX86::GenerateIndirectExitCode( u32 num_instructions, CIndirectExitMap * p_map )
{

	MOVI(ECX_CODE, num_instructions);
	CALL( CCodeLabel( CPU_UpdateCounter ) );

	CCodeLabel		no_target( NULL );
	CJumpLocation	jump_to_next_fragment( GenerateBranchIfNotSet( &gCPUState.StuffToDo, no_target ) );

	CCodeLabel		exit_dynarec( GetAssemblyBuffer()->GetLabel() );
	// New return address is in gCPUState.TargetPC
	MOV_REG_MEM( EAX_CODE, &gCPUState.TargetPC );
	MOV_MEM_REG( &gCPUState.CurrentPC, EAX_CODE );
	SetVar8( &gCPUState.Delay, NO_DELAY );
	
	// No need to call CPU_SetPC(), as this is handled by CFragment when we exit

	RET();

	// gCPUState.StuffToDo == 0, try to jump to the indirect target
	PatchJumpLong( jump_to_next_fragment, GetAssemblyBuffer()->GetLabel() );

	MOVI( ECX_CODE, reinterpret_cast< u32 >( p_map ) );
	MOV_REG_MEM( EDX_CODE, &gCPUState.TargetPC );
	CALL( CCodeLabel( IndirectExitMap_Lookup ) );

	// If the target was not found, exit
	TEST( EAX_CODE, EAX_CODE );
	JELong( exit_dynarec );

	JMP_REG( EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::GenerateExceptionHander( ExceptionHandlerFn p_exception_handler_fn, std::vector< CJumpLocation > & exception_handler_jumps )
{
	CCodeLabel exception_handler( GetAssemblyBuffer()->GetLabel() );

	CALL( CCodeLabel( p_exception_handler_fn ) );
	RET();

	for( std::vector< CJumpLocation >::iterator it = exception_handler_jumps.begin(); it != exception_handler_jumps.end(); ++it )
	{
		CJumpLocation	jump( *it );
		PatchJumpLong( jump, exception_handler );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::PreEmitCachedRegCheck( const SRegisterStats & stats )
{
	if (!mSpCachedInESI && stats.RegUpdates[REG_sp] == mSetSpPostUpdate)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  Setting sp in ESI after update %d", stats.RegUpdates[REG_sp] ) );
		
		mSpCachedInESI = true;
		
		// Calculate SP address and store in esi:
		GenerateReadWriteAddress(REG_sp, 0, 0, false, ~0, false, false, ESI_CODE);
	}

	// Record how many updates we've had
	for (u32 i = 0; i < NUM_MIPS_REGISTERS; i++)
	{
		gWriteCheck[i] = stats.RegUpdates[i];
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised)
{
	// If the operation was NOT optimised, and our cached register was written to, 
	// then we have to reload the value:
	if (!optimised)
	{
		// Also, check hi regs! gRegisterStatus.IsHiUnknown( i )
		for (u32 i = 0; i < NUM_MIPS_REGISTERS; i++)
		{
			if (gWriteCheck[i] != stats.RegUpdates[i])
			{
				DPF( DEBUG_DYNREC, DSPrintf( "  Reg %s was updated", RegNames[i] ) );

				// See if the lo value is invalid
				if (gRegisterStatus.GetCachedReg( i ) != INVALID_CODE)
				{
					gRegisterStatus.MarkAsValid( i, false );
				}

				// See if the hi value we have is invalid
				if (!gRegisterStatus.IsHiUnknown( i ))
				{
					gRegisterStatus.MarkHiAsUnknown( i, true );
					gRegisterStatus.MarkHiAsDirty( i, false );
				}
			}
		}
	}
}


//*****************************************************************************
//	If the register is cached, that register is used.
//	Otherwise the specified scratch register is used.
//*****************************************************************************
EIntelReg	CCodeGeneratorX86::GetMIPSLoRegisterNoLoad(u32 mreg, EIntelReg scratch_to_use)
{
	EIntelReg ireg( gRegisterStatus.GetCachedReg( mreg ) );
	if(ireg == INVALID_CODE)
	{
		ireg = scratch_to_use;
	}

	return ireg;
}

//*****************************************************************************
//	If the register is cached, that register is used.
//	Otherwise the specified scratch register is used.
//*****************************************************************************
EIntelReg	CCodeGeneratorX86::GetMIPSLoRegisterAndLoad(u32 mreg, EIntelReg scratch_to_use)
{
	EIntelReg ireg( gRegisterStatus.GetCachedReg( mreg ) );

	bool need_load( false );
	
	if(ireg == INVALID_CODE)
	{
		ireg = scratch_to_use;
		need_load = true;
	}
	else
	{
		if (gRegisterStatus.IsValid( mreg ))	
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  ++ Using cached value for %s...\\/", RegNames[mreg] ) );
		}
		else
		{
			gRegisterStatus.MarkAsValid( mreg, true );
			need_load = true;
			DPF( DEBUG_DYNREC, DSPrintf( "  ++ Cached value for %s is invalid - loading...\\/", RegNames[mreg] ) );
		}
	}

	if(need_load)
	{
		if(mreg == REG_r0)
		{
			DPF(DEBUG_DYNREC, "  ++ Clearing reg for r0 load...\\/");		
			// We have to use MOV and not XOR to avoid setting the flags
			//XOR(ireg, ireg);											
			MOVI(ireg, 0);
		}
		else
		{
			MOV_REG_MEM( ireg, GetRegisterAddressLo(mreg) );	
		}
	}

	return ireg;
}

//*****************************************************************************
//	If the MIPS register is cached, flag it as being dirty
//	Otherwise write the result back immediately
//*****************************************************************************
void	CCodeGeneratorX86::UpdateMIPSReg( u32 mreg, EIntelReg src, EUpdateRegOptions options, EIntelReg scratch )
{
	EIntelReg	icached( gRegisterStatus.GetCachedReg( mreg ) );

	if(icached != INVALID_CODE)
	{
		if(icached != src)
		{
			//
			//	If the registers are different, it means we're discarding the old value
			//
			MOV( icached, src );
		}
		gRegisterStatus.MarkAsDirty( mreg, true );
		gRegisterStatus.MarkAsValid( mreg, true );
	}
	else
	{
		MOV_MEM_REG( GetRegisterAddressLo( mreg ), src );
	}

	//
	//	Sign extend etc
	//
	switch(options)
	{
	case URO_HI_LEAVE:
		break;
	case URO_HI_SIGN_EXTEND:
		if(src != scratch)
		{
			MOV(scratch, src);
		}

		SARI(scratch, 31);
		StoreMIPSHi(mreg, scratch);
		break;
	case URO_HI_CLEAR:
		SetMIPSHi( mreg, 0 );
		break;
	case URO_HI_SET:
		SetMIPSHi( mreg, 0xffffffff );
		break;
	default:
		DAEDALUS_ASSERTMSG( "Unhandled option" );
		NODEFAULT;
		break;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::LoadMIPSLo(EIntelReg ireg, u32 mreg)	
{
	EIntelReg iCachedReg( gRegisterStatus.GetCachedReg( mreg ) );
	
	if (iCachedReg != INVALID_CODE)
	{	
		if (gRegisterStatus.IsValid( mreg ))	
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  ++ Using cached value for %s...\\/", RegNames[mreg] ) );
		}
		else
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  ++ Cached value for %s is invalid - loading...\\/", RegNames[mreg] ) );
			MOV_REG_MEM(iCachedReg, GetRegisterAddressLo(mreg) );	
			gRegisterStatus.MarkAsValid( mreg, true );
		}

		MOV(ireg, iCachedReg);											
	}																																		
	else if (mreg == REG_r0)											
	{
		DPF(DEBUG_DYNREC, "  ++ Clearing reg for r0 load...\\/");		
		// We have to use MOV and not XOR to avoid setting the flags
		//XOR(ireg, ireg);											
		MOVI(ireg, 0);
	}
	else																
	{
		MOV_REG_MEM(ireg, GetRegisterAddressLo(mreg) );	
	}
}

//*****************************************************************************
// TODO - if the reg is cached, check signed/zero/set flags?!?
//*****************************************************************************
void CCodeGeneratorX86::LoadMIPSHi(EIntelReg ireg, u32 mreg)
{
	if (mreg == REG_r0)
	{
		DPF(DEBUG_DYNREC, "  ++ Clearing reg for r0 load...\\/");
		//XOR(ireg, ireg);
		MOVI(ireg, 0);
	}
	else
	{
		if (gRegisterStatus.IsHiUnknown( mreg ))
		{
			MOV_REG_MEM(ireg, GetRegisterAddressHi(mreg) );
		}
		else
		{
			DPF( DEBUG_DYNREC, DSPrintf( "  ++ We could optimise hi reg load here (loading 0x%08x)!!!", gRegisterStatus.GetHiValue( mreg ) ) );
			MOVI(ireg, gRegisterStatus.GetHiValue( mreg ));
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::StoreMIPSLo(u32 mreg, EIntelReg ireg)
{
	EIntelReg iCachedReg = gRegisterStatus.GetCachedReg( mreg );
	
	if (iCachedReg != INVALID_CODE)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  ++ Updating cached value for %s...\\/", RegNames[mreg] ) );
		MOV(iCachedReg, ireg);
		gRegisterStatus.MarkAsDirty( mreg, true );
		gRegisterStatus.MarkAsValid( mreg, true );
	}
	else
	{
		MOV_MEM_REG( GetRegisterAddressLo(mreg), (ireg) );
	}
}

//*****************************************************************************
// TODO - Keep record of signed/zero/set flags?
//*****************************************************************************
void CCodeGeneratorX86::StoreMIPSHi(u32 mreg, EIntelReg ireg)
{
	MOV_MEM_REG(GetRegisterAddressHi(mreg), ireg );

	// We know longer know the contents of the high register...
	if (!gRegisterStatus.IsHiUnknown( mreg ))
	{
		DPF(DEBUG_DYNREC, "  ++ We've successfully avoided a hiwriteback!");
	}
	gRegisterStatus.MarkHiAsUnknown( mreg, true );
	gRegisterStatus.MarkHiAsDirty( mreg, false );
	
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::SetMIPSLo(u32 mreg, u32 data)
{
	EIntelReg iCachedReg = gRegisterStatus.GetCachedReg( mreg );
	
	if (iCachedReg != INVALID_CODE)
	{
		DPF( DEBUG_DYNREC, DSPrintf( "  ++ Setting cached value for %s...\\/", RegNames[mreg] ) );
		MOVI(iCachedReg, data);
		gRegisterStatus.MarkAsDirty( mreg, true );
		gRegisterStatus.MarkAsValid( mreg, true );
	}
	else
	{
		MOVI_MEM( GetRegisterAddressLo(mreg), data);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorX86::SetMIPSHi(u32 mreg, u32 data)
{
	//MOVI_MEM( GetRegisterAddressHi(mreg), (data));

	DPF(DEBUG_DYNREC, "  ++ We could stall writeback here!!!");
	gRegisterStatus.MarkHiAsUnknown( mreg, false );
	gRegisterStatus.SetHiValue( mreg, data );
	gRegisterStatus.MarkHiAsDirty( mreg, true );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::SetVar( u32 * p_var, u32 value )
{
	MOVI_MEM( p_var, value );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::SetVar8( u32 * p_var, u8 value )
{
	MOVI_MEM8( p_var, value );
}

//*****************************************************************************
// this is really, really, really ugly - however until the dynarec support routines get better, this cannot be easily split
//*****************************************************************************
void	CCodeGeneratorX86::GenerateReadWriteAddress(u32 base, u32 offset, u32 xor_val, bool write, u32 rt, bool baseimm, bool base_already_loaded, EIntelReg destreg)
{
	if(!base_already_loaded)
	{
		if(baseimm)
		{
			MOVI(destreg, base);
		}
		else if(base > 31) // memory location
		{
			MOV_REG_MEM(destreg, (void*)base);
		}
		else
		{
			LoadMIPSLo(destreg, base);
		}
	}

	// Add offset
	if (offset != 0)
	{
		ADDI(destreg, offset);
	}
	
	MOV(EDX_CODE, destreg);

	// it's here to improve parallelism
	// but maybe it would be better to put it after the JS
	if (xor_val)
	{
		XOR_I32(destreg, xor_val);
	}

	SHRI(EDX_CODE, 18);

	// ADD EAX, [g_ReadAddressPointerLookupTable + EDX*4]
	ADD_REG_MEM_IDXx4( destreg, write ? g_WriteAddressPointerLookupTable : g_ReadAddressPointerLookupTable, EDX_CODE );

	mpSecondary->PadTo16Bytes();

	// JS special
	static const u32	JS_LEN( 6 );
	CCodeLabel			secondary_label( mpSecondary->GetLabel() );
	CJumpLocation		jump_location( mpPrimary->GetJumpLocation() );
	JS( jump_location.GetOffset( secondary_label ) - JS_LEN );

	bool was_rt_valid = gRegisterStatus.IsValid( rt );

	if(rt != u32(~0))
	{
		LoadMIPSLo(EDX_CODE, rt);

		// MOV [EAX], EDX
		MOV_MEM_BASE_REG( destreg, EDX_CODE );
	}

	SetAssemblyBuffer( mpSecondary );

	// MOV ECX, address
	if(baseimm)
	{
		MOVI( ECX_CODE, (u32)base );
	}
	else if(base > 31) // memory location
	{
		MOV_REG_MEM( ECX_CODE, (void *)base );
	}
	else
	{
		// similar to LoadMIPSLo
		EIntelReg iCachedReg = gRegisterStatus.GetCachedReg( base );

		if (iCachedReg != INVALID_CODE)
		{	
			MOV( ECX_CODE, iCachedReg );
		}																																		
		else if (base == REG_r0)		
		{
			// XOR ECX, ECX
			XOR( ECX_CODE, ECX_CODE );
		}
		else																
		{
			MOV_REG_MEM( ECX_CODE, (u8*)&gGPR[0] + (base * 8) );
		}
	}

	if (rt != u32(~0))
	{
		MOV( EAX_CODE, EDX_CODE );

		// MOV EDX, rt
		// similar to LoadMIPSLo
		EIntelReg iCachedReg = gRegisterStatus.GetCachedReg( rt );

		if (iCachedReg != INVALID_CODE)
		{
			if (!was_rt_valid)	
			{
				MOV_REG_MEM( iCachedReg, (u8*)&gGPR[0] + (rt * 8) );
				gRegisterStatus.MarkAsValid( rt, true );
			}

			MOV( EDX_CODE, iCachedReg );
		}																																		
		else if (rt == REG_r0)		
		{
			// XOR EDX, EDX
			XOR( EDX_CODE, EDX_CODE );
		}
		else																
		{
			MOV_REG_MEM( EDX_CODE, (u8*)&gGPR[0] + (rt * 8) );
		}
	}

	if(offset)
	{
		ADDI( ECX_CODE, offset );
	}

	if(xor_val)
	{
		XOR_I8( ECX_CODE, u8( xor_val ) );
	}

	// CALL DWORD PTR [g_ReadAddressLookupTable + EAX/EDX*4]
	if(rt != u32(~0))
	{
		CALL_MEM_PLUS_REGx4( write ? (void*)g_WriteAddressValueLookupTable : (void*)g_ReadAddressLookupTable, EAX_CODE );
	}
	else
	{
		CALL_MEM_PLUS_REGx4( write ? (void*)g_WriteAddressLookupTable : (void*)g_ReadAddressLookupTable, EDX_CODE );
	}

	MOV( destreg, EAX_CODE );

	// JMP after_pointer_null
	JMPLong( mpPrimary->GetLabel() );

	SetAssemblyBuffer( mpPrimary );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::CleanupForExit()
{
	// Flush all FP regs that are currently cached
	SR_FP_FlushAllRegs();
	
	if (mSetSpPostUpdate != u32(~0))
		POP(ESI_CODE);

	FlushMipsRegs();
	RestoreIntelRegs();
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::FlushMipsRegs()
{
	//
	//	Flush the Hi parts of the registers first
	//	(index from 1 as we skip $r0)
	//
	for (u32 i = 1; i < NUM_MIPS_REGISTERS; i++)
	{
		if (gRegisterStatus.IsHiDirty( i ) && !gRegisterStatus.IsHiUnknown( i ))
		{
			// Writeback
			DPF( DEBUG_DYNREC, DSPrintf( "Writing back info for %s/hi: 0x%08x", RegNames[i], gRegisterStatus.GetHiValue( i ) ) );

			MOVI_MEM( GetRegisterAddressHi(i), gRegisterStatus.GetHiValue( i ));

			gRegisterStatus.MarkHiAsDirty( i, false );
		}
	}

	// Pop regs from lowest to highest...
	for (u32 iReg = 0; iReg < NUM_X86_REGISTERS; iReg++)
	{
		u32 mips_reg = gIntelRegUsageMap[iReg];

		// If we have cached a register, we may need to flush it
		if (mips_reg != INTEL_REG_UNUSED)
		{
			if (gRegisterStatus.IsDirty( mips_reg ))
			{
				if (gRegisterStatus.IsValid( mips_reg ))
				{
					DPF( DEBUG_DYNREC, DSPrintf( "Cached reg %s is dirty: flushing", RegNames[mips_reg] ) );
					
					MOV_MEM_REG( GetRegisterAddressLo(mips_reg), (EIntelReg)iReg );

					// We could update the modcount here?
				}
				gRegisterStatus.MarkAsDirty( mips_reg, false );
			}
		}
	}
}


//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::RestoreIntelRegs()
{
	// Pop regs from lowest to highest...
	for (u32 i = 0; i < NUM_X86_REGISTERS; i++)
	{
		u32 mips_reg = gIntelRegUsageMap[i];

		// If we have cached a register, we may need to flush it
		if (mips_reg != INTEL_REG_UNUSED)
		{
			gRegisterStatus.MarkAsValid( mips_reg, false );

			// Always restore!
			POP( EIntelReg( i ) );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBlockEnd()
{
	MOV_REG_MEM(EAX_CODE, &gCPUState.Delay);
	CMPI(EAX_CODE, NO_DELAY);			// Delay set?
	
	JE(10);
	// Just let the core handle the jump for now...change from DO_DELAY to EXEC_DELAY
	MOVI_MEM( &gCPUState.Delay, EXEC_DELAY );		// 10 bytes
}

namespace
{
	//*****************************************************************************
	// Returns non-zero if we should continue
	//*****************************************************************************
	u32 CheckStuffCallback( u32 num_ops )
	{
		// Increment count register
		CPUEvent * p_first_event( CPU_GetFirstEvent() );

		u32		num_cycles( num_ops * COUNTER_INCREMENT_PER_OP );
		if (p_first_event->mCount > s32(num_cycles) )
		{
			g_qwCPR[0][C0_COUNT]._u32[0] = g_qwCPR[0][C0_COUNT]._u32[0] + num_cycles;
			p_first_event->mCount-= num_cycles;

			return gCPUState.StuffToDo;		// Don't return 0 - return flags if they're set
		}
		else
		{
			num_cycles = p_first_event->mCount - 1;
			
			g_qwCPR[0][C0_COUNT]._u32[0] = g_qwCPR[0][C0_COUNT]._u32[0] + num_cycles;
			p_first_event->mCount = 1;

			return 1;
		}

	}

}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops )
{
	// If NO_DELAY, skip all this, and return
	MOV_REG_MEM(EAX_CODE, &gCPUState.Delay);
	TEST(EAX_CODE, EAX_CODE);			// NO_DELAY == 0, 
	
	JE(5+5+2+2+2+
		10+5+
		10);			// ->branch_not_taken

	MOVI(ECX_CODE, num_ops);			// 5
	MOVI(EAX_CODE, reinterpret_cast< u32 >( CheckStuffCallback ));	// Copy p to eax - 5
	CALL_REG( EAX_CODE );					// 2
	TEST(EAX_CODE, EAX_CODE);				// 2
	JNE(10+5);		//-> stuff_to_do				// 2

	// Set delay instruction to NO_DELAY - have we handled the op?
	MOVI_MEM( &gCPUState.Delay, NO_DELAY );		// 10 bytes
	// Branch to the top of the code..
	JMPLong( CCodeLabel(p_target->mpFunctionPointer) );		// 5 bytes

// stuff_to_do
	// If gCPUState.Delay was DO_DELAY, we come here
	MOVI_MEM( &gCPUState.Delay, EXEC_DELAY );		// 10 bytes
// branch_not_taken
	// No delay, because the jump was not taken
	// If gCPUState.Delay == NO_DELAY, then we jump here
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBranchHandler( CJumpLocation branch_handler_jump, RegisterSnapshotHandle snapshot )
{
	CCodeLabel current_label( GetAssemblyBuffer()->GetLabel() );

	PatchJumpLong( branch_handler_jump, current_label );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchAlways( CCodeLabel target )
{
	return JMPLong( target );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchIfSet( const u32 * p_var, CCodeLabel target )
{
	MOV_REG_MEM( EAX_CODE, p_var );
	TEST( EAX_CODE, EAX_CODE );

	return JNELong( target );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchIfNotSet( const u32 * p_var, CCodeLabel target )
{
	MOV_REG_MEM( EAX_CODE, p_var );
	TEST( EAX_CODE, EAX_CODE );

	return JELong( target );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchIfEqual32( const u32 * p_var, u32 value, CCodeLabel target )
{
	CMP_MEM32_I32( p_var, value );
	
	return JELong( target );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchIfEqual8( const u32 * p_var, u8 value, CCodeLabel target )
{
	CMP_MEM32_I8( p_var, value );
	
	return JELong( target );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchIfNotEqual32( const u32 * p_var, u32 value, CCodeLabel target )
{
	CMP_MEM32_I32( p_var, value );
	
	return JNELong( target );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateBranchIfNotEqual8( const u32 * p_var, u8 value, CCodeLabel target )
{
	CMP_MEM32_I8( p_var, value );
	
	return JNELong( target );
}

//*****************************************************************************
//	Generates instruction handler for the specified op code.
//	Returns a jump location if an exception handler is required
//*****************************************************************************
CJumpLocation	CCodeGeneratorX86::GenerateOpCode( u32 address, bool branch_delay_slot, OpCode op_code, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	bool	need_pc( R4300_InstructionHandlerNeedsPC( op_code ) );

	if( !need_pc )
	{
		address = 0;
	}

	if( branch_delay_slot )
	{
		SetVar8( &gCPUState.Delay, EXEC_DELAY );
	}

	GenerateGenericR4300( address, op_code, R4300_GetInstructionHandler( op_code ) );

	CJumpLocation	exception_handler;
	CCodeLabel		no_target( NULL );

	if( need_pc )
	{
		exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, no_target );
	}

	// Check whether we want to invert the status of this branch
	if( p_branch != NULL )
	{
		//
		// Check if the branch has been taken
		//
		if( p_branch->Direct )
		{
			if( p_branch->ConditionalBranchTaken )
			{
				*p_branch_jump = GenerateBranchIfNotEqual8( &gCPUState.Delay, DO_DELAY, no_target );
			}
			else
			{
				*p_branch_jump = GenerateBranchIfEqual8( &gCPUState.Delay, DO_DELAY, no_target );
			}
		}
		else
		{
			// XXXX eventually just exit here, and skip default exit code below
			if( p_branch->Eret )
			{
				*p_branch_jump = GenerateBranchAlways( no_target );
			}
			else
			{
				*p_branch_jump = GenerateBranchIfNotEqual32( &gCPUState.TargetPC, p_branch->TargetAddress, no_target );
			}
		}
	}
	else
	{
		if( branch_delay_slot )
		{
			SetVar8( &gCPUState.Delay, NO_DELAY );
		}
	}

	
	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateGenericR4300( u32 address, OpCode op_code, CPU_Instruction p_instruction )
{
	// XXXX Flush all fp registers before a generic call

	if( address != 0 )
	{
		SetVar( &gCPUState.CurrentPC, address );
	}
	
	// Call function - __fastcall
	MOVI(ECX_CODE, op_code._u32);
	CALL( CCodeLabel( p_instruction ) );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction pF )
{
	// Flush all fp registers before a generic call
	if (optimise)
	{
		FlushMipsRegs();		// Valid flag updated after op "executed"
		SR_FP_FlushAllRegs();
	}

	// Set up op_code

	// __fastcall
	MOVI(ECX_CODE, op_code._u32);
	
	// __cdecl
	//MOVI(ECX_CODE, op_code._u32);
	//PUSH( ECX_CODE );

	// Lkb: I think it's better to only inline functions when dwOptimiseLevel
	// is greater than zero. This avoid inlining for functions that are only
	// ever hit once, saving buffer space and "one-off" overheads.
	if (!optimise)
	{
		// Call function
		CALL( CCodeLabel( pF ) );
	}
	else
	{
//		if(!DynaRec_Emit_Inline_Function(this, (void*)pF))
		{
			// Call function
			CALL( CCodeLabel( pF ) );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::ExecuteSpeedHack( CCodeLabel speed_hack )
{
	CALL( speed_hack );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMFHI( EN64Reg rd )
{
	//gGPR[dwRD] = g_qwMultHI;

	MOV_REG_MEM(EAX_CODE, (BYTE *)&g_qwMultHI + 0);
	MOV_REG_MEM(EDX_CODE, (BYTE *)&g_qwMultHI + 4);

	StoreMIPSLo(rd, EAX_CODE);
	StoreMIPSHi(rd, EDX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMFLO( EN64Reg rd )
{
	//gGPR[dwRD] = g_qwMultLO;

	MOV_REG_MEM(EAX_CODE, (BYTE *)&g_qwMultLO + 0);
	MOV_REG_MEM(EDX_CODE, (BYTE *)&g_qwMultLO + 4);

	StoreMIPSLo(rd, EAX_CODE);
	StoreMIPSHi(rd, EDX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMULTU( EN64Reg rs, EN64Reg rt )
{
	//u64 dwResult = ((u64)(u32)gGPR[dwRS]) * ((u64)(u32)gGPR[dwRT]);
	//g_qwMultLO = (s64)(s32)(dwResult);
	//g_qwMultHI = (s64)(s32)(dwResult >> 32);

	// We can do the following optimisations:
	//  
	// If neither is cached, load rs, mul rs, [rt]
	// If rs is cached, and rt is not, do mul rs, [rt]
	// If rt is cached, and rs is not, do mul rt, [rs]
	// If both are cached, need to store one and use that value

	// Optimise for in-place addition
	EIntelReg iCachedRegRS = gRegisterStatus.GetCachedReg( rs );
	EIntelReg iCachedRegRT = gRegisterStatus.GetCachedReg( rt );

	bool bCachedRS = iCachedRegRS != INVALID_CODE;
	bool bCachedRT = iCachedRegRT != INVALID_CODE;

	if (!bCachedRT)
	{
		LoadMIPSLo(EAX_CODE, rs);

		// mul eax, [dwRT]
		MUL_EAX_MEM( GetRegisterAddressLo(rt) );
	}
	else if (!bCachedRS && bCachedRT)
	{
		LoadMIPSLo(EAX_CODE, rt);
		// mul eax, [dwRS]
		MUL_EAX_MEM( GetRegisterAddressLo(rs) );
	}
	else		// Both cached
	{
		LoadMIPSLo(EAX_CODE, rs);
		LoadMIPSLo(iCachedRegRT, rt);			// Ensure register is loaded
		// mul eax, dwRT
		MUL(iCachedRegRT);
	}

	MOV_MEM_REG((u8 *)&g_qwMultLO + 0, EAX_CODE);
	SARI(EAX_CODE, 31);		
	MOV_MEM_REG((u8 *)&g_qwMultLO + 4, EAX_CODE);

	MOV_MEM_REG((u8 *)&g_qwMultHI + 0, EDX_CODE);
	SARI(EDX_CODE, 31);		
	MOV_MEM_REG((u8 *)&g_qwMultHI + 4, EDX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateADDU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[dwRD] = (s64)((s32)gGPR[dwRS] + (s32)gGPR[dwRT]);

	EIntelReg	iDstReg( GetMIPSLoRegisterNoLoad( rd, EAX_CODE ) );
	EIntelReg	iSrc1Reg;
	EIntelReg	iSrc2Reg;

	// Optimise ADD eax, dword ptr [uncached reg]
	// could do "lea		dst, [eax+eax]?

	if (rd == rt)
	{
		// Swap register order so that we avoid clobbering rd with the MOV below
		iSrc1Reg = GetMIPSLoRegisterAndLoad( rt, EAX_CODE );
		iSrc2Reg = GetMIPSLoRegisterAndLoad( rs, EDX_CODE );
	}
	else
	{
		iSrc1Reg = GetMIPSLoRegisterAndLoad( rs, EAX_CODE );
		iSrc2Reg = GetMIPSLoRegisterAndLoad( rt, EDX_CODE );
	}

	if(iSrc1Reg != iDstReg)
	{
		MOV(iDstReg, iSrc1Reg);
	}

	ADD(iDstReg, iSrc2Reg);
	UpdateMIPSReg( rd, iDstReg, URO_HI_SIGN_EXTEND, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSUBU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[dwRD] = (s64)((s32)gGPR[dwRS] - (s32)gGPR[dwRT]);

	LoadMIPSLo(EAX_CODE, rs);
	LoadMIPSLo(EDX_CODE, rt);

	SUB(EAX_CODE, EDX_CODE);

	UpdateMIPSReg( rd, EAX_CODE, URO_HI_SIGN_EXTEND, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateAND( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[dwRD] = gGPR[dwRS] & gGPR[dwRT];

	if (rs == REG_r0 || rt == REG_r0)
	{
		// This doesn't seem to happen much - is it worth optimising?
		// One or both registers are zero - we can just clear the destination register
		SetMIPSLo(rd, 0);
		SetMIPSHi(rd, 0);
	}
	else
	{
		// Both registers non zero - perform operation as normal
		LoadMIPSLo(EAX_CODE, rs);
		LoadMIPSLo(EDX_CODE, rt);
		AND(EAX_CODE, EDX_CODE);
		StoreMIPSLo(rd, EAX_CODE);

		LoadMIPSHi(EDX_CODE, rt);	// Put EDX first to allow pipelining with above
		LoadMIPSHi(EAX_CODE, rs);
		AND(EAX_CODE, EDX_CODE);
		StoreMIPSHi(rd, EAX_CODE);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[rd] = gGPR[rs] | gGPR[rt];

	if (rs == 0 && rt == 0)
	{
		// Both registers are zero - we can just clear the destination register
		SetMIPSLo(rd, 0);
		SetMIPSHi(rd, 0);
	}
	else if (rs == 0)
	{
		// This case rarely seems to happen...
		// As RS is zero, the OR is just a copy of RT to RD.
		LoadMIPSLo(EAX_CODE, rt);
		LoadMIPSHi(EDX_CODE, rt);
		StoreMIPSLo(rd, EAX_CODE);
		StoreMIPSHi(rd, EDX_CODE);
	}
	else if (rt == 0)
	{
		// As RT is zero, the OR is just a copy of RS to RD.
		LoadMIPSLo(EAX_CODE, rs);
		LoadMIPSHi(EDX_CODE, rs);
		StoreMIPSLo(rd, EAX_CODE);
		StoreMIPSHi(rd, EDX_CODE);
	}
	else
	{
		// Both registers non zero - perform operation as normal
		// Can we just do OR(EAX_CODE, lohalf(rt)) here????
		LoadMIPSLo(EAX_CODE, rs);
		LoadMIPSLo(EDX_CODE, rt);
		OR(EAX_CODE, EDX_CODE);
		StoreMIPSLo(rd, EAX_CODE);

		LoadMIPSHi(EDX_CODE, rt);	// Put EDX first to allow pipelining with above
		LoadMIPSHi(EAX_CODE, rs);
		OR(EAX_CODE, EDX_CODE);
		StoreMIPSHi(rd, EAX_CODE);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateXOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[dwRD] = gGPR[dwRS] ^ gGPR[dwRT];

	// We could special case RS or RT being zero (or the same), but this never seems to happen
	// Both registers non zero - perform operation as normal
	LoadMIPSLo(EAX_CODE, rs);
	LoadMIPSLo(EDX_CODE, rt);
	XOR(EAX_CODE, EDX_CODE);
	StoreMIPSLo(rd, EAX_CODE);

	LoadMIPSHi(EDX_CODE, rt);	// Put EDX first to allow pipelining with above
	LoadMIPSHi(EAX_CODE, rs);
	XOR(EAX_CODE, EDX_CODE);
	StoreMIPSHi(rd, EAX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateNOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[rd] = ~(gGPR[rs] | gGPR[rt]);
	// Optimise for second source being 0 (just ~ first)
	if (rt == 0)
	{
		LoadMIPSLo(EAX_CODE, rs);
		LoadMIPSHi(EDX_CODE, rs);	// Use EDX to allow pipelining with above
		NOT(EAX_CODE);
		NOT(EDX_CODE);
		StoreMIPSLo(rd, EAX_CODE);
		StoreMIPSHi(rd, EDX_CODE);
	}
	else
	{
		// Both registers non zero - perform operation as normal
		LoadMIPSLo(EAX_CODE, rs);
		LoadMIPSLo(EDX_CODE, rt);
		OR(EAX_CODE, EDX_CODE);
		NOT(EAX_CODE);
		StoreMIPSLo(rd, EAX_CODE);

		LoadMIPSHi(EDX_CODE, rt);	// Put EDX first to allow pipelining with above
		LoadMIPSHi(EAX_CODE, rs);
		OR(EAX_CODE, EDX_CODE);
		NOT(EAX_CODE);
		StoreMIPSHi(rd, EAX_CODE);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSLT( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	// Cast to ints to ensure sign is taken into account
	//if ((s64)gGPR[rs] < (s64)gGPR[rt]) {
	//	gGPR[rd] = 1;
	//} else {
	//	gGPR[rd] = 0;
	//}

	if (rs == REG_r0)
	{
		// TODO - CMPI with 0, not CMP with ecx
		XOR(ECX_CODE, ECX_CODE);		// RS == 0
		LoadMIPSHi(EDX_CODE, rt);

		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETL(EAX_CODE);					// 3 bytes

		// RS/lo is still 0
		LoadMIPSLo(EDX_CODE, rt);		// 
		
		JNE(0x5);						// Next two instructions are 5 bytes

		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETB(EAX_CODE);					// 3 bytes
	}
	else if (rt == REG_r0)
	{
		// TODO - CMPI with 0?
		LoadMIPSHi(ECX_CODE, rs);
		XOR(EDX_CODE, EDX_CODE);		// RT == 0

		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETL(EAX_CODE);					// 3 bytes

		LoadMIPSLo(ECX_CODE, rs);		// See comments on SLT for why we load before branch
		// RT is still 0

		JNE(0x5);						// Next two instructions are 5 bytes

		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETB(EAX_CODE);					// 3 bytes
	}
	else
	{
		LoadMIPSHi(ECX_CODE, rs);
		LoadMIPSHi(EDX_CODE, rt);

		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETL(EAX_CODE);					// 3 bytes

		LoadMIPSLo(ECX_CODE, rs);		// See comments on SLT for why we load before branch
		LoadMIPSLo(EDX_CODE, rt);		// 
		
		JNE(0x5);						// Next two instructions are 5 bytes

		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETB(EAX_CODE);					// 3 bytes
	}
	// Store result in RD
	UpdateMIPSReg( rd, EAX_CODE, URO_HI_CLEAR, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSLTU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	// Code is the same as above, but uses SETB for both tests
	
	if (rs == REG_r0)
	{
		XOR(ECX_CODE, ECX_CODE);		// RS is zero!
		LoadMIPSHi(EDX_CODE, rt);

		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETB(EAX_CODE);					// 3 bytes
		
		// ECX is still 0
		LoadMIPSLo(EDX_CODE, rt);				// See comments on SLT for why we load before branch
		
		JNE(0x5);						// Next two instructions are 5 bytes

		CMP(ECX_CODE, EDX_CODE);	// 2 bytes
		SETB(EAX_CODE);				// 3 bytes
	}
	else if (rt == REG_r0)
	{
		LoadMIPSHi(ECX_CODE, rs);
		XOR(EDX_CODE, EDX_CODE);		// RT is zero

		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETB(EAX_CODE);					// 3 bytes
		
		LoadMIPSLo(ECX_CODE, rs);		// See comments on SLT for why we load before branch
		// EDX is still 0
		
		JNE(0x5);						// Next two instructions are 5 bytes

		CMP(ECX_CODE, EDX_CODE);	// 2 bytes
		SETB(EAX_CODE);				// 3 bytes
	}
	else
	{
		LoadMIPSHi(ECX_CODE, rs);
		LoadMIPSHi(EDX_CODE, rt);

		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMP(ECX_CODE, EDX_CODE);		// 2 bytes
		SETB(EAX_CODE);					// 3 bytes
		
		LoadMIPSLo(ECX_CODE, rs);		// See comments on SLT for why we load before branch
		LoadMIPSLo(EDX_CODE, rt);		//
		
		JNE(0x5);						// Next two instructions are 5 bytes

		CMP(ECX_CODE, EDX_CODE);	// 2 bytes
		SETB(EAX_CODE);				// 3 bytes
	}
		
	// Store result in RD
	UpdateMIPSReg( rd, EAX_CODE, URO_HI_CLEAR, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateADDIU( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	//gGPR[rt] = (s64)gGPR[rs] + (s64)(s32)wData;

	// We can do the following optimisations:
	//  If rs is reg0, we just copy the immediate data to regRT
	//  If wData is 0, we can copy regRS to regRT

	s32		data( immediate );

	if (rs == REG_r0)
	{
		// The source register is zero, so we only have to consider the immediate value
		SetMIPSLo(rt, data);
		if (data >= 0)
		{
			SetMIPSHi(rt, 0);		// Set top bits to zero
		}
		else
		{
			SetMIPSHi(rt, 0xFFFFFFFF);	// Set top bits to -1
		}
	}
	else
	{
		EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rt, EAX_CODE ) );
		EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rs, iDstReg ) );		// Try to avoid MOV below if possible

		if(iSrcReg != iDstReg)
		{
			MOV( iDstReg, iSrcReg );
		}
		ADDI(iDstReg, data);
		UpdateMIPSReg( rt, iDstReg, URO_HI_LEAVE, EAX_CODE );

		// XXXX this is incorrect - this is DADDIU behaviour
		// Finally perform high-word addition
		// WARN - the ADCI will only work if the flags have not been modified by a store above!
		LoadMIPSHi(EDX_CODE, rs);
		if (data >= 0 )
			ADCI(EDX_CODE, 0);
		else
			ADCI(EDX_CODE, -1);		// 0xFF

		StoreMIPSHi(rt, EDX_CODE);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateANDI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	u32	data( immediate );
	//gGPR[rt] = gGPR[rs] & (u64)wData;
	// Note that top half is always set to 0...

	// Optimise for ANDI inplace
	EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rt, EAX_CODE ) );
	EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rs, iDstReg ) );		// Try to avoid MOV below if possible

	if(iSrcReg != iDstReg)
	{
		MOV( iDstReg, iSrcReg );
	}
	ANDI(iDstReg, data);
	UpdateMIPSReg( rt, iDstReg, URO_HI_CLEAR, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateORI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	u32	data( immediate );
	//gGPR[rt] = gGPR[rs] | (u64)wData;

	if (rs == REG_r0)					// Setting to immediate data
	{
		// g_qwGRP[rt] = (u64)dwData
		DPF(DEBUG_DYNREC, "  ++ ORI0: Setting value");
		SetMIPSLo(rt, data);
		SetMIPSHi(rt, 0);
	}
	else								// "Normal" operation source/dest different
	{
		EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rt, EAX_CODE ) );
		EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rs, iDstReg ) );		// Try to avoid MOV below if possible

		if(iSrcReg != iDstReg)
		{
			MOV( iDstReg, iSrcReg );
		}
		ORI(iDstReg, data);
		UpdateMIPSReg( rt, iDstReg, URO_HI_LEAVE, EAX_CODE );

		//
		//	If the source/dest are different we need to copy the hi bits
		//
		if(rt != rs)
		{
			// Copy top half, as regs are different			
			LoadMIPSHi(EAX_CODE, rs);
			StoreMIPSHi(rt, EAX_CODE);
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateXORI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	u32	data( immediate );
	//gGPR[rt] = gGPR[rs] ^ (u64)wData;
	EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rt, EAX_CODE ) );
	EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rs, iDstReg ) );		// Try to avoid MOV below if possible

	if(iSrcReg != iDstReg)
	{
		MOV( iDstReg, iSrcReg );
	}
	XOR_I32(iDstReg, data);
	UpdateMIPSReg( rt, iDstReg, URO_HI_LEAVE, EAX_CODE );

	// Copy high bits to rt (i.e. xoring with 0)
	// Don't bother if source/dest are the same e.g. XORI s1,s1,0xFFFF
	if (rt != rs)
	{
		LoadMIPSHi(EDX_CODE, rs);
		StoreMIPSHi(rt, EDX_CODE);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLUI( EN64Reg rt, s16 immediate )
{
	s32		data( ((s32)immediate) << 16 );

	//gGPR[rt] = (s64)(s32)(wData<<16);

	// TODO: If a lui 0x80xx or 0xa0xx is performed,
	// flag the register as "ram pointer". This will
	// let us optimise other regs but SP in Store 
	// ops

	// Check for next op being ADDIU or ADDI (common way of setting entire register
	//if (dwNextOp.op == ADDIU || dwNextOp.op == ADDI) {

	SetMIPSLo(rt, data);
	if (data >= 0)
	{
		SetMIPSHi(rt, 0);			// Clear tops bits
	}
	else
	{
		SetMIPSHi(rt, 0xFFFFFFFF);	// Set top bits
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSLTI( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	s32	data( immediate );

	// Cast to ints to ensure sign is taken into account
	//gGPR[rt] = ((s64)gGPR[rs] < (s64)(s32)wData ? 1: 0);

	// TODO optimise for when cmp reg is cached/valid (compare direct)

	if (data >= 0)
	{
		// data is positive
		LoadMIPSHi(EDX_CODE, rs);
		XOR(EAX_CODE, EAX_CODE);		// Clear dlo so that SETcc AL works
		CMPI(EDX_CODE, 0x00000000);	// 6 bytes
		SETL(EAX_CODE);						// 3 bytes

		LoadMIPSLo(EDX_CODE, rs);	// Doesn't affect flags, but we must load before branch to preserve our internal record of the status
		
		JNE(0x9);							// Next two instructions are 9 bytes
		CMPI(EDX_CODE, data);		// 6 bytes
		SETB(EAX_CODE);						// 3 bytes
	}
	else 
	{
		// data is negative
		LoadMIPSHi(EDX_CODE, rs);
		XOR(EAX_CODE, EAX_CODE);		// Clear EAX so that SETcc AL works
		CMPI(EDX_CODE, 0xFFFFFFFF);	// 6 bytes
		SETL(EAX_CODE);						// 3 bytes
		
		LoadMIPSLo(EDX_CODE, rs);	// See comments above
		
		JNE(0x9);							// Next two instructions are 9 bytes
		CMPI(EDX_CODE, data);		// 6 bytes
		SETB(EAX_CODE);						// 3 bytes
	}

	// Store result in RT
	UpdateMIPSReg( rt, EAX_CODE, URO_HI_CLEAR, EAX_CODE );

}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSLTIU( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	s32	data( immediate );

	// Cast to ints to ensure sign is taken into account
	//gGPR[rt] = ((s64)gGPR[rs] < (s64)(s32)wData ? 1: 0);

	LoadMIPSHi(EDX_CODE, rs);
	if (data >= 0) {
		// data is positive
		XOR(EAX_CODE, EAX_CODE);	// Clear dlo so that SETcc AL works
		CMPI(EDX_CODE, 0x00000000);	// 6 bytes
		SETB(EAX_CODE);						// 3 bytes
		
		LoadMIPSLo(EDX_CODE, rs);	// See comments on SLT for why we load before branch
		
		JNE(0x9);							// Next two instructions are 9 bytes
		CMPI(EDX_CODE, data);		// 6 bytes
		SETB(EAX_CODE);						// 3 bytes
	} else {
		// data is negative
		XOR(EAX_CODE, EAX_CODE);		// Clear EAX so that SETcc AL works
		CMPI(EDX_CODE, 0xFFFFFFFF);	// 6 bytes
		SETB(EAX_CODE);						// 3 bytes
		
		LoadMIPSLo(EDX_CODE, rs);	// See comments on SLT for why we load before branch
		
		JNE(0x9);							// Next two instructions are 9 bytes
		CMPI(EDX_CODE, data);		// 6 bytes
		SETB(EAX_CODE);						// 3 bytes
	}

	// Store result in RT
	UpdateMIPSReg( rt, EAX_CODE, URO_HI_CLEAR, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSLL( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[rd] = (s64)(s32)((((u32)gGPR[rt]) << (sa)) & 0xFFFFFFFF);

	EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rd, EAX_CODE ) );
	EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rt, iDstReg ) );		// Try to avoid MOV below if possible

	if(iSrcReg != iDstReg)
	{
		MOV( iDstReg, iSrcReg );
	}

	SHLI( iDstReg, u8( sa ) );
	UpdateMIPSReg( rd, iDstReg, URO_HI_SIGN_EXTEND, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSRL( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[rd] = (s64)(s32)(((u32)gGPR[rt]) >> sa);

	EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rd, EAX_CODE ) );
	EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rt, iDstReg ) );		// Try to avoid MOV below if possible

	if(iSrcReg != iDstReg)
	{
		MOV( iDstReg, iSrcReg );
	}

	SHRI(iDstReg, u8(sa));
	UpdateMIPSReg( rd, iDstReg, URO_HI_CLEAR, EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSRA( EN64Reg rd, EN64Reg rt, u32 sa )
{
	// Need to check that this correctly works with sign
	//gGPR[rd] = (s64)(s32)(((s32)gGPR[rt]) >> sa);

	EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rd, EAX_CODE ) );
	EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rt, iDstReg ) );		// Try to avoid MOV below if possible

	if(iSrcReg != iDstReg)
	{
		MOV( iDstReg, iSrcReg );
	}

	SARI(iDstReg, u8(sa));

	// Is sign extension necessary? Top bits will ALWAYS be same as before??
	UpdateMIPSReg(rd, iDstReg, URO_HI_SIGN_EXTEND, EAX_CODE);

}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLB( EN64Reg rt, EN64Reg base, s32 offset )
{
	// TODO: If the base is SP, and we have SP cached, use ESI instead:

	GenerateReadWriteAddress(base, offset, 3, false);
	
	// mov al, byte ptr [eax]
	// movsx	eax, al
	MOV8_REG_MEM_BASE( EAX_CODE, EAX_CODE );
	MOVSX8( EAX_CODE, EAX_CODE );

	UpdateMIPSReg(rt, EAX_CODE, URO_HI_SIGN_EXTEND, EAX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLBU( EN64Reg rt, EN64Reg base, s32 offset )
{

	// TODO: If the base is SP, and we have SP cached, use ESI instead:

	GenerateReadWriteAddress(base, offset, 3, false);
	
	// mov al, byte ptr [eax]
	// and		eax, 0x00FF	- Mask off top bits!
	MOV8_REG_MEM_BASE( EAX_CODE, EAX_CODE );
	AND_EAX( 0x000000FF );

	// Top bits always 0
	UpdateMIPSReg(rt, EAX_CODE, URO_HI_CLEAR, EAX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLH( EN64Reg rt, EN64Reg base, s32 offset )
{
	// TODO: If the base is SP, and we have SP cached, use ESI instead:

	//DWORD dwAddress = (u32)((s32)gGPR[dwBase] + (s32)wOffset);
	//gGPR[dwRT] = (s64)(s16)Read16Bits(dwAddress);

	GenerateReadWriteAddress(base, offset, 2, false);

	// mov ax, word ptr [eax]
	// movsx	eax, ax
	MOV16_REG_MEM_BASE( EAX_CODE, EAX_CODE );
	MOVSX16( EAX_CODE, EAX_CODE );

	UpdateMIPSReg(rt, EAX_CODE, URO_HI_SIGN_EXTEND, EAX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLHU( EN64Reg rt, EN64Reg base, s32 offset )
{
	// TODO: If the base is SP, and we have SP cached, use ESI instead:

	//DWORD dwAddress = (u32)((s32)gGPR[dwBase] + (s32)wOffset);
	//gGPR[dwRT] = (u64)(u16)Read16Bits(dwAddress);

	GenerateReadWriteAddress(base, offset, 2, false);

	// mov ax, word ptr [eax]
	// and		eax, 0xFFFF	- Mask off top bits!
	MOV16_REG_MEM_BASE( EAX_CODE, EAX_CODE );
	AND_EAX( 0x0000FFFF );

	// Top bits always 0
	UpdateMIPSReg(rt, EAX_CODE, URO_HI_CLEAR, EAX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLW( EN64Reg rt, EN64Reg base, s16 _offset )
{
	s32			offset( _offset );		// XXXX
	EIntelReg	ibase;
	s32			base_offset;

	// If the base is SP, and we have SP cached, use ESI instead:
	if (base == REG_sp && mSpCachedInESI)
	{
		DPF(DEBUG_DYNREC, " Using cached reg in ESI");
		ibase = ESI_CODE;
		base_offset = offset;
	}
	else
	{
		GenerateReadWriteAddress(base, offset, 0, false);

		ibase = EAX_CODE;
		base_offset = 0;
	}

	// mov eax, dword ptr [base+base_offset]
	MOV_REG_MEM_BASE_OFFSET( EAX_CODE, ibase, base_offset );
	UpdateMIPSReg(rt, EAX_CODE, URO_HI_SIGN_EXTEND, EAX_CODE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateLWC1( u32 ft, EN64Reg base, s32 offset )
{
	// Simulate SR_FP stuff
	SR_FP_LWC1(ft);

	EIntelReg	ibase;
	s32			base_offset;

	// If the base is SP, and we have SP cached, use ESI instead:
	if (base == REG_sp && mSpCachedInESI)
	{
		DPF(DEBUG_DYNREC, " Using cached reg in ESI");
		ibase = ESI_CODE;
		base_offset = offset;
	}
	else
	{
		GenerateReadWriteAddress(base, offset, 0, false);

		ibase = EAX_CODE;
		base_offset = 0;
	}

	MOV_REG_MEM_BASE_OFFSET( EAX_CODE, ibase, base_offset );
	MOV_MEM_REG(&g_qwCPR[1][ft], EAX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSB( EN64Reg rt, EN64Reg base, s32 offset )
{
	// TODO - Try and use cached base?
	// This is different from SW(), because for the most part we don't care about
	// Writes to system registers!

	GenerateReadWriteAddress(base, offset, 3, true);

	EIntelReg	iSrcReg( GetMIPSLoRegisterAndLoad( rt, EDX_CODE ) );

	// MOV8 requires EAX/ECX/EDX/EBX. We have to copy to a scratch reg to get around this
	if(!IsValidMov8Reg( iSrcReg ))
	{
		MOV( EDX_CODE, iSrcReg );
		iSrcReg = EDX_CODE;
	}

	// mov byte ptr [eax], dl
	MOV8_MEM_BASE_REG( EAX_CODE, iSrcReg );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSH( EN64Reg rt, EN64Reg base, s32 offset )
{
	// TODO - Try and use cached base?
	// This is different from SW(), because for the most part we don't care about
	// Writes to system registers!

	GenerateReadWriteAddress(base, offset, 2, true);

	// EAX hold memory address to write to
	// Get value to write
	EIntelReg	iSrcReg( GetMIPSLoRegisterAndLoad( rt, EDX_CODE ) );

	// mov word ptr [eax], dx
	MOV16_MEM_BASE_REG( EAX_CODE, iSrcReg );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSW( EN64Reg rt, EN64Reg base, s32 offset )
{
	// We only optimise write using sp as the base (pushes onto the stack)
	// The reason for this is that other bases may point to hardware regs,
	// and so we may need to Update memory following the write
	if (base == REG_sp && mSpCachedInESI)
	{
		DPF(DEBUG_DYNREC, " Using cached reg in ESI");

		EIntelReg	iSrcReg( GetMIPSLoRegisterAndLoad( rt, EAX_CODE ) );

		MOV_MEM_BASE_OFFSET_REG( ESI_CODE, offset, iSrcReg );
	}
	else
	{
		// This code is different from SWC1, because we want to call MemoryUpdate after
		// the call to update system registers, copy memory etc
		GenerateReadWriteAddress(base, offset, 0, true, rt);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSWC1( u32 ft, EN64Reg base, s32 offset )
{
	/*
	DWORD dwAddress = (u32)((s32)gGPR[dwBase] + (s32)wOffset);
	Write32Bits(dwAddress, LoadFPR_Word(dwFT));
	*/

	// Simulate SR_FP stuff
	SR_FP_SWC1(ft);

	EIntelReg	ibase;
	s32			base_offset;

	// If the base is SP, and we have SP cached, use ESI instead:
	if (base == REG_sp && mSpCachedInESI)
	{
		DPF(DEBUG_DYNREC, " Using cached reg in ESI");

		ibase = ESI_CODE;
		base_offset = offset;
	}
	else
	{
		// This is different from SW(), because for the most part we don't care about
		// Writes to system registers!
		
		// Get base (this could be cached!)
		GenerateReadWriteAddress(base, offset, 0, true);

		ibase = EAX_CODE;
		base_offset = 0;
	}

	MOV_REG_MEM(EDX_CODE, &g_qwCPR[1][ft]);
	MOV_MEM_BASE_OFFSET_REG( ibase, base_offset, EDX_CODE );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMFC1( EN64Reg rt, u32 fs )
{
	// Simulate SR_FP stuff
	SR_FP_MFC1(fs);

	// MFC1 in the manual says this is a sign-extended result
	//gGPR[dwRT] = (s64)(s32)LoadFPR_Word(dwFS);

	// Get value from fp reg
	MOV_REG_MEM(EAX_CODE, &g_qwCPR[1][fs]);

	StoreMIPSLo(rt, EAX_CODE);
	SARI(EAX_CODE, 31);		
	StoreMIPSHi(rt, EAX_CODE);

}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMTC1( u32 fs, EN64Reg rt )
{
	// Manual says top bits undefined after load
	//StoreFPR_Word(dwFS, (s32)gGPR[dwRT]);

	// Simulate SR_FP stuff
	SR_FP_MTC1(fs);

	EIntelReg iReg = gRegisterStatus.GetCachedReg( rt );

	// Optimisation - avoid reg->reg copy by moving value directly!
	if (iReg != INVALID_CODE && gRegisterStatus.IsValid( rt ))
	{
		DPF(DEBUG_DYNREC, "  ++ Using value direct from cahced reg...\\/");
		MOV_MEM_REG(&g_qwCPR[1][fs], iReg);
	}
	else
	{
		LoadMIPSLo(EAX_CODE, rt);
		MOV_MEM_REG(&g_qwCPR[1][fs], EAX_CODE);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCFC1( EN64Reg rt, u32 fs )
{
	// Only defined for reg 0 or 31
	if (fs == 0 || fs == 31)
	{
		// Just do Lo load/store - high bits not valid
		//gGPR[dwRT] = (s64)(s32)g_qwCCR[1][dwFS];

		EIntelReg iDstReg( GetMIPSLoRegisterNoLoad( rt, EAX_CODE ) );

		MOV_REG_MEM(iDstReg, &g_qwCCR[1][fs]);	
		UpdateMIPSReg( rt, iDstReg, URO_HI_LEAVE, EAX_CODE );
	}
	else
	{
		// Ignore!
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCTC1( u32 fs, EN64Reg rt )
{
	// Only defined for reg 0 or 31
	if (fs == 0 || fs == 31)
	{
		// Just do Lo load/store - high bits not valid
		//g_qwCCR[1][dwFS] = gGPR[dwRT];
		EIntelReg iSrcReg( GetMIPSLoRegisterAndLoad( rt, EAX_CODE ) );
	
		MOV_MEM_REG(&g_qwCCR[1][fs], iSrcReg);	
	}
	else
	{
		// Ignore!
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBEQ( EN64Reg rs, EN64Reg rt, u32 target )
{
	// Perform the comparison - use cached regs if possible
	if (rs == rt)
	{
		DPF(DEBUG_DYNREC, "  ++ Optimizing for branch always taken...\\/");		
		// Branch is always taken
	}
	else if (rt == REG_r0)	// dwRS rarely seems to be 0
	{
		DPF(DEBUG_DYNREC, "  ++ Comparing against constant 0...\\/");		
		EIntelReg iReg( GetMIPSLoRegisterAndLoad( rs, EAX_CODE ) );

		CMPI(iReg, 0);
		JNE(10 + 10);
	}
	else
	{
		EIntelReg iRegS( GetMIPSLoRegisterAndLoad( rs, EAX_CODE ) );
		EIntelReg iRegT( GetMIPSLoRegisterAndLoad( rt, EDX_CODE ) );

		CMP(iRegS, iRegT);
		JNE(10 + 10);
	}

	//XXXX Optimise with SetVar8
	//CPU_TakeBranch( target );				// Warning- this macro might change in future!
	SetVar( &gCPUState.TargetPC, target );	// 10 bytes
	SetVar( &gCPUState.Delay, DO_DELAY );	// 10 bytes
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBNE( EN64Reg rs, EN64Reg rt, u32 target )
{
	// Perform the comparison - use cached regs if possible
	if (rt == REG_r0)	// rs rarely seems to be 0
	{
		DPF(DEBUG_DYNREC, "  ++ Comparing against constant 0...\\/");		

		EIntelReg iReg( GetMIPSLoRegisterAndLoad( rs, EAX_CODE ) );
		CMPI(iReg, 0);
	}
	else
	{
		EIntelReg iRegS( GetMIPSLoRegisterAndLoad( rs, EAX_CODE ) );
		EIntelReg iRegT( GetMIPSLoRegisterAndLoad( rt, EDX_CODE ) );

		CMP(iRegS, iRegT);
	}
	
	JE(10 + 10);
	//XXXX Optimise with SetVar8
	//CPU_TakeBranch( target );				// Warning- this macro might change in future!
	SetVar( &gCPUState.TargetPC, target );	// 10 bytes
	SetVar( &gCPUState.Delay, DO_DELAY );	// 10 bytes

	// Jump here if branch not taken
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBLEZ( EN64Reg rs, u32 target )
{
	// Perform the comparison - use cached regs if possible
	EIntelReg iReg( GetMIPSLoRegisterAndLoad( rs, EAX_CODE ) );

	CMPI(iReg, 0);
	
	JG(10 + 10);	// Jump if NOT LE
	//XXXX Optimise with SetVar8
	//CPU_TakeBranch( target );				// Warning- this macro might change in future!
	SetVar( &gCPUState.TargetPC, target );	// 10 bytes
	SetVar( &gCPUState.Delay, DO_DELAY );	// 10 bytes

	// Jump here if branch not taken
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateBGTZ( EN64Reg rs, u32 target )
{
	// Perform the comparison - use cached regs if possible
	EIntelReg iReg( GetMIPSLoRegisterAndLoad( rs, EAX_CODE ) );

	CMPI(iReg, 0);
	
	JLE(10 + 10);	// Jump if NOT G
	//CPU_TakeBranch( target );				// Warning- this macro might change in future!
	//XXXX Optimise with SetVar8
	SetVar( &gCPUState.TargetPC, target );	// 10 bytes
	SetVar( &gCPUState.Delay, DO_DELAY );	// 10 bytes

	// Jump here if branch not taken
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateADD_S( u32 fd, u32 fs, u32 ft )
{
	SR_FP_ADD(fd, fs, ft, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSUB_S( u32 fd, u32 fs, u32 ft )
{
	SR_FP_SUB(fd, fs, ft, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMUL_S( u32 fd, u32 fs, u32 ft )
{
	SR_FP_MUL(fd, fs, ft, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateDIV_S( u32 fd, u32 fs, u32 ft )
{
	SR_FP_DIV(fd, fs, ft, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSQRT_S( u32 fd, u32 fs )
{
	SR_FP_SQRT(fd, fs, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMOV_S( u32 fd, u32 fs )
{
	SR_FP_MOV(fd, fs, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateNEG_S( u32 fd, u32 fs )
{
	SR_FP_NEG(fd, fs, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCompare_S( u32 fs, u32 ft, u16 flag )
{
	SR_FP_C(fs, ft, flag, FP_SINGLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateADD_D( u32 fd, u32 fs, u32 ft )
{
	SR_FP_ADD(fd, fs, ft, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSUB_D( u32 fd, u32 fs, u32 ft )
{
	SR_FP_SUB(fd, fs, ft, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMUL_D( u32 fd, u32 fs, u32 ft )
{
	SR_FP_MUL(fd, fs, ft, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateDIV_D( u32 fd, u32 fs, u32 ft )
{
	SR_FP_DIV(fd, fs, ft, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateSQRT_D( u32 fd, u32 fs )
{
	SR_FP_SQRT(fd, fs, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateMOV_D( u32 fd, u32 fs )
{
	SR_FP_MOV(fd, fs, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateNEG_D( u32 fd, u32 fs )
{
	SR_FP_NEG(fd, fs, FP_DOUBLE);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCompare_D( u32 fs, u32 ft, u16 flag )
{
	SR_FP_C(fs, ft, flag, FP_DOUBLE);
}


//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCVT_D_S( u32 fd, u32 fs )
{
	SR_FP_S_CVT_D(fd, fs);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCVT_W_S( u32 fd, u32 fs )
{
	SR_FP_S_CVT_W(fd, fs);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCVT_D_W( u32 fd, u32 fs )
{
	SR_FP_W_CVT_D(fd, fs);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCVT_S_W( u32 fd, u32 fs )
{
	SR_FP_W_CVT_S(fd, fs);
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorX86::GenerateCVT_S_D( u32 fd, u32 fs )
{
	SR_FP_D_CVT_S(fd, fs);
}
