/*
Copyright (C) 2001 StrmnNrmn

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

#include "stdafx.h"
#include "TextureCache.h"
#include "ConvertImage.h"
#include "DebugDisplayList.h"
#include "Core\Memory.h"

#include "PixelFormat.h"

#include "DaedMathUtil.h"

#include "RDP.h"

using namespace PixelFormats;

namespace
{

const u8 TwoToEight[4] =
{
	0x00,		// 00 -> 00 00 00 00
	0x55,		// 01 -> 01 01 01 01
	0xaa,		// 10 -> 10 10 10 10
	0xff		// 11 -> 11 11 11 11
};

const u8 FourToEight[16] = 
{
	0x00, 0x11, 0x22, 0x33,
	0x44, 0x55, 0x66, 0x77,
	0x88, 0x99, 0xaa, 0xbb,
	0xcc, 0xdd, 0xee, 0xff
};

template< u32 Size >
struct SByteswapInfo;

template<> struct SByteswapInfo< 1 >
{
	enum { Fiddle = 3 };
};
template<> struct SByteswapInfo< 2 >
{
	enum { Fiddle = 2 };
};
template<> struct SByteswapInfo< 4 >
{
	enum { Fiddle = 0 };
};

}

//*****************************************************************************
//
//*****************************************************************************
template < typename OutT >
struct SConvertGeneric
{
typedef void (*ConvertRowFunction)( OutT * p_dst, const u8 * p_src_base, u32 offset, u32 width );


static void ConvertGeneric( CTexture *pSurf,
							const TextureInfo & ti,
							ConvertRowFunction swapped_fn,
							ConvertRowFunction unswapped_fn )
{
	DrawInfo dst;
		
	if (!pSurf->StartUpdate(&dst))
		return;

	OutT *				p_dst( reinterpret_cast< OutT * >( dst.pSurface ) );
	
#if RDP_EMULATE_TMEM
	u32					blah( ti.GetSize() == G_IM_SIZ_32b ? 2 : 1 );

	const u8 *			p_src_base( gTextureMemory );
	u32					base_offset( ti.TmemAddress << 3 );
	u32					src_pitch( gRDPTiles[ gTextureTile ].line * 8 * blah );
	const bool			force_swapped( true );
#else
	const u8 *			p_src_base( g_pu8RamBase );
	u32					base_offset( ti.GetBaseOffset() );
	u32					src_pitch( ti.Pitch );
	const bool			force_swapped( false );
#endif

	if (force_swapped || ti.IsSwapped())
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
			{
				unswapped_fn( p_dst, p_src_base, base_offset, ti.Width );
			}
			else
			{
				swapped_fn( p_dst, p_src_base, base_offset, ti.Width );
			}

			base_offset += src_pitch;
			p_dst = reinterpret_cast< OutT * >( (u8*)p_dst + dst.Pitch );
		}
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			unswapped_fn( p_dst, p_src_base, base_offset, ti.Width );

			base_offset += src_pitch;
			p_dst = reinterpret_cast< OutT * >( (u8*)p_dst + dst.Pitch );
		}
	}
	
	pSurf->EndUpdate(&dst);
}

};

typedef void (*ConvertPalettisedRowFunction)( Psp::Pf8888 * p_dst, const u8 * p_src_base, u32 offset, u32 width, const u16 * p_palette );

void ConvertGenericPalettised( CTexture *pSurf, const TextureInfo & ti, ConvertPalettisedRowFunction swapped_fn, ConvertPalettisedRowFunction unswapped_fn )
{
	DrawInfo dst;
		
	if (!pSurf->StartUpdate(&dst))
		return;

	Psp::Pf8888 *		p_dst( reinterpret_cast< Psp::Pf8888 * >( dst.pSurface ) );

#if RDP_EMULATE_TMEM
	u32					blah( ti.GetSize() == G_IM_SIZ_32b ? 2 : 1 );

	const u8 *			p_src_base( gTextureMemory );
	u32					base_offset( ti.TmemAddress << 3 );
	u32					src_pitch( gRDPTiles[ gTextureTile ].line * 8 * blah );
	const bool			force_swapped( true );
#else
	const u8 *			p_src_base( g_pu8RamBase );
	u32					base_offset( ti.GetBaseOffset() );
	u32					src_pitch( ti.Pitch );
	const bool			force_swapped( false );
#endif

	const u16 *			p_palette( reinterpret_cast< const u16 * >( &gTextureMemory[ ti.TmemPalAddress << 3 ] ) );

	if (force_swapped || ti.IsSwapped())
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
			{
				unswapped_fn( p_dst, p_src_base, base_offset, ti.Width, p_palette );
			}
			else
			{
				swapped_fn( p_dst, p_src_base, base_offset, ti.Width, p_palette );
			}

			base_offset += src_pitch;
			p_dst = reinterpret_cast< Psp::Pf8888 * >( (u8*)p_dst + dst.Pitch );
		}
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			unswapped_fn( p_dst, p_src_base, base_offset, ti.Width, p_palette );

			base_offset += src_pitch;
			p_dst = reinterpret_cast< Psp::Pf8888 * >( (u8*)p_dst + dst.Pitch );
		}
	}
	
	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
//
//*****************************************************************************
template < typename InT >
struct SConvert
{
	enum { Fiddle = SByteswapInfo< sizeof( InT ) >::Fiddle };

	//
	//	This routine converts from any format which is > 1 byte to any Psp format.
	//
	template < typename OutT, u32 F > 
	static inline void ConvertRow( OutT * p_dst, const u8 * p_src_base, u32 offset, u32 width )
	{
		DAEDALUS_DL_ASSERT( IsAligned( offset, sizeof( InT ) ), "Offset should be correctly aligned" );

		for (u32 x = 0; x < width; x++)
		{
			InT	colour( *reinterpret_cast< const InT * >( &p_src_base[offset ^ F] ) );

			p_dst[x] = convertPixelFormat< OutT, InT >( colour );

			offset += sizeof( InT );
		}
	}

	template < typename OutT >
	static inline void ConvertTextureT( CTexture *pSurf, const TextureInfo & ti )
	{
		SConvertGeneric< OutT >::ConvertGeneric( pSurf, ti, ConvertRow< OutT, 0x4 | Fiddle >, ConvertRow< OutT, Fiddle > );
	}

	static void ConvertTexture( CTexture *pSurf, const TextureInfo & ti )
	{
		switch( pSurf->GetTexture()->GetFormat() )
		{
		case TexFmt_5650:	ConvertTextureT< Psp::Pf5650 >( pSurf, ti ); return;
		case TexFmt_5551:	ConvertTextureT< Psp::Pf5551 >( pSurf, ti ); return;
		case TexFmt_4444:	ConvertTextureT< Psp::Pf4444 >( pSurf, ti ); return;
		case TexFmt_8888:	ConvertTextureT< Psp::Pf8888 >( pSurf, ti ); return;
		}

		DAEDALUS_DL_ERROR( "Unhandled format" );
	}
};

//*****************************************************************************
//
//*****************************************************************************
struct SConvertIA4
{
	enum { Fiddle = 0x3 };

	template < typename OutT, u32 F >
	static inline void ConvertRow( OutT * p_dst, const u8 * p_src_base, u32 offset, u32 width ) 
	{
		DAEDALUS_DL_ASSERT( (width&1) == 0, "Expecting an even width" );

		// Do two pixels at a time
		for (u32 x = 0; x < width; x+=2)
		{
			u8 b = p_src_base[offset ^ F];

			// Even
			p_dst[x + 0] = OutT( TwoToEight[(b & 0xc0) >> 6],
								 TwoToEight[(b & 0xc0) >> 6],
								 TwoToEight[(b & 0xc0) >> 6],
								 TwoToEight[(b & 0x30) >> 4] );	
			// Odd
			p_dst[x + 1] = OutT( TwoToEight[(b & 0x0c) >> 2],
								 TwoToEight[(b & 0x0c) >> 2],
								 TwoToEight[(b & 0x0c) >> 2],
								 TwoToEight[(b & 0x03)     ] );
			offset++;
		}
	}

	template < typename OutT >
	static inline void ConvertTextureT( CTexture *pSurf, const TextureInfo & ti )
	{
		SConvertGeneric< OutT >::ConvertGeneric( pSurf, ti, ConvertRow< OutT, 0x4 | Fiddle >, ConvertRow< OutT, Fiddle > );
	}

	static void ConvertTexture( CTexture *pSurf, const TextureInfo & ti )
	{
		switch( pSurf->GetTexture()->GetFormat() )
		{
		case TexFmt_5650:	ConvertTextureT< Psp::Pf5650 >( pSurf, ti ); return;
		case TexFmt_5551:	ConvertTextureT< Psp::Pf5551 >( pSurf, ti ); return;
		case TexFmt_4444:	ConvertTextureT< Psp::Pf4444 >( pSurf, ti ); return;
		case TexFmt_8888:	ConvertTextureT< Psp::Pf8888 >( pSurf, ti ); return;
		}

		DAEDALUS_DL_ERROR( "Unhandled format" );
	}
};

//*****************************************************************************
//
//*****************************************************************************
struct SConvertI4
{
	enum { Fiddle = 0x3 };

	template< typename OutT, u32 F >
	static inline void ConvertRow( OutT * p_dst, const u8 * p_src_base, u32 offset, u32 width )
	{
		DAEDALUS_DL_ASSERT( (width&1) == 0, "Expecting an even width" );

		// Do two pixels at a time
		for ( u32 x = 0; x+1 < width; x+=2 )
		{
			u8 b = p_src_base[offset ^ F];

			// Even
			p_dst[x + 0] = OutT( FourToEight[(b & 0xF0)>>4],
								 FourToEight[(b & 0xF0)>>4],
								 FourToEight[(b & 0xF0)>>4],
								 FourToEight[(b & 0xF0)>>4] );	
			// Odd
			p_dst[x + 1] = OutT( FourToEight[(b & 0x0F)],
								 FourToEight[(b & 0x0F)],
								 FourToEight[(b & 0x0F)],
								 FourToEight[(b & 0x0F)] );

			offset++;
		}
	}

	template < typename OutT >
	static inline void ConvertTextureT( CTexture *pSurf, const TextureInfo & ti )
	{
		SConvertGeneric< OutT >::ConvertGeneric( pSurf, ti, ConvertRow< OutT, 0x4 | Fiddle >, ConvertRow< OutT, Fiddle > );
	}

	static void ConvertTexture( CTexture *pSurf, const TextureInfo & ti )
	{
		switch( pSurf->GetTexture()->GetFormat() )
		{
		case TexFmt_5650:	ConvertTextureT< Psp::Pf5650 >( pSurf, ti ); return;
		case TexFmt_5551:	ConvertTextureT< Psp::Pf5551 >( pSurf, ti ); return;
		case TexFmt_4444:	ConvertTextureT< Psp::Pf4444 >( pSurf, ti ); return;
		case TexFmt_8888:	ConvertTextureT< Psp::Pf8888 >( pSurf, ti ); return;
		}

		DAEDALUS_DL_ERROR( "Unhandled format" );
	}
};

//*****************************************************************************
//
//*****************************************************************************
template< typename PalT, u32 F > void ConvertCI4_Row( Psp::Pf8888 * p_dst, const u8 * p_src_base, u32 offset, u32 width, const u16 * p_palette )
{
	const PalT *	p_n64pal( reinterpret_cast< const PalT * >( p_palette ) );

	for (u32 x = 0; x < width; x+=2)
	{
		u8 b = p_src_base[offset ^ F];

		u8 bhi = (b&0xf0)>>4;
		u8 blo = (b&0x0f);

		p_dst[ x + 0 ] = Psp::Pf8888::Make( p_n64pal[ bhi ^ 0x1 ] );	// Remember palette is in different endian order!
		p_dst[ x + 1 ] = Psp::Pf8888::Make( p_n64pal[ blo ^ 0x1 ] );

		offset++;
	}
}

//*****************************************************************************
//
//*****************************************************************************
template< typename PalT, u32 F > void ConvertCI8_Row( Psp::Pf8888 * p_dst, const u8 * p_src_base, u32 offset, u32 width, const u16 * p_palette )
{
	const PalT *	p_n64pal( reinterpret_cast< const PalT * >( p_palette ) );

	for (u32 x = 0; x < width; x++)
	{
		u8 b = p_src_base[offset ^ F];

		p_dst[ x ] = Psp::Pf8888::Make( p_n64pal[ b ^ 0x1 ] );	// Remember palette is in different endian order!

		offset++;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void ConvertRGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	SConvert< N64::Pf5551 >::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
//
//*****************************************************************************
void ConvertRGBA32(CTexture *pSurf, const TextureInfo & ti)
{
	// Did have Fiddle of 8 here, pretty sure this was wrong (should have been 4)
	SConvert< N64::Pf8888 >::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
// E.g. Dear Mario text
// Copy, Score etc
//*****************************************************************************
void ConvertIA4(CTexture *pSurf, const TextureInfo & ti)
{
	SConvertIA4::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
// E.g Mario's head textures
//*****************************************************************************
void ConvertIA8(CTexture *pSurf, const TextureInfo & ti)
{
	SConvert< N64::PfIA8 >::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
// E.g. camera's clouds, shadows
//*****************************************************************************
void ConvertIA16(CTexture *pSurf, const TextureInfo & ti)
{
	SConvert< N64::PfIA16 >::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
// Used by MarioKart
//*****************************************************************************
void ConvertI4(CTexture *pSurf, const TextureInfo & ti)
{
	SConvertI4::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
// Used by MarioKart
//*****************************************************************************
void ConvertI8(CTexture *pSurf, const TextureInfo & ti)
{
	SConvert< N64::PfI8 >::ConvertTexture( pSurf, ti );
}

//*****************************************************************************
// Used by Starfox intro
//*****************************************************************************
void ConvertCI4_RGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DAEDALUS_DL_ASSERT( pSurf->GetTexture()->GetFormat() == TexFmt_8888, "This is not an 8888 format texture" );

	ConvertGenericPalettised( pSurf, ti, ConvertCI4_Row< N64::Pf5551, 0x4 | 0x3 >, ConvertCI4_Row< N64::Pf5551, 0x3 > );
}

//*****************************************************************************
// Used by Starfox intro
//*****************************************************************************
void ConvertCI4_IA16(CTexture *pSurf, const TextureInfo & ti)
{
	DAEDALUS_DL_ASSERT( pSurf->GetTexture()->GetFormat() == TexFmt_8888, "This is not an 8888 format texture" );

	ConvertGenericPalettised( pSurf, ti, ConvertCI4_Row< N64::PfIA16, 0x4 | 0x3 >, ConvertCI4_Row< N64::PfIA16, 0x3 > );
}

//*****************************************************************************
// Used by MarioKart for Cars etc
//*****************************************************************************
void ConvertCI8_RGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DAEDALUS_DL_ASSERT( pSurf->GetTexture()->GetFormat() == TexFmt_8888, "This is not an 8888 format texture" );

	ConvertGenericPalettised( pSurf, ti, ConvertCI8_Row< N64::Pf5551, 0x4 | 0x3 >, ConvertCI8_Row< N64::Pf5551, 0x3 > );
}

//*****************************************************************************
// Used by MarioKart for Cars etc
//*****************************************************************************
void ConvertCI8_IA16(CTexture *pSurf, const TextureInfo & ti)
{
	DAEDALUS_DL_ASSERT( pSurf->GetTexture()->GetFormat() == TexFmt_8888, "This is not an 8888 format texture" );

	ConvertGenericPalettised( pSurf, ti, ConvertCI8_Row< N64::PfIA16, 0x4 | 0x3 >, ConvertCI8_Row< N64::PfIA16, 0x3 > );
}


