/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


#ifndef NATIVETEXTURE_H_
#define NATIVETEXTURE_H_

#include "DaedRefCounted.h"

#include "TextureFormat.h"

class c32;

class CNativeTexture : public daedalus::CRefCounted
{
		CNativeTexture( u32 w, u32 h, ETextureFormat texture_format );
		~CNativeTexture();
	public:

		static	void					CorrectDimensions( u32 width, u32 height, u32 & adjusted_width, u32 & adjusted_height );

		static	CNativeTexture *		Create( u32 width, u32 height, ETextureFormat texture_format );
		static	CNativeTexture *		CreateFromRaw( const char * p_filename, u32 width, u32 height, ETextureFormat texture_format );
		static	CNativeTexture *		CreateFromAlpha( const char * p_filename, u32 width, u32 height, ETextureFormat texture_format );
		static	CNativeTexture *		CreateFromPng( const char * p_filename, ETextureFormat texture_format );

		bool							LoadPngData( const char * p_filename );

		void							InstallTexture();

		void							SetData( void * p_data, u32 w, u32 h );
		void							SetDataForceColour( void * p_data, u32 w, u32 h, u32 mask, u32 colour );			// Set the data, but replace RGB with the specified colour (Alpha remains)
		CNativeTexture *				Recolour( c32 colour ) const;

		u32								GetWidth() const				{ return mWidth; }
		u32								GetHeight() const				{ return mHeight; }
		u32								GetCorrectedWidth() const		{ return mCorrectedWidth; }
		u32								GetCorrectedHeight() const		{ return mCorrectedHeight; }
		u32								GetStride() const;
		ETextureFormat					GetFormat() const		{ return mTextureFormat; }

		u32								GetBytesRequired() const;

		u32								GetVideoMemoryUsage() const;
		u32								GetSystemMemoryUsage() const;

	private:
		ETextureFormat		mTextureFormat;
		u32					mWidth;
		u32					mHeight;
		u32					mCorrectedWidth;
		u32					mCorrectedHeight;
		void *				mpData;
		bool				mIsVidMem;
		bool				mIsSwizzled;
};

#endif	// NATIVETEXTURE_H_
