/*
Copyright (C) 2001 Lkb

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "PngUtil.h"
#include "TextureFormat.h"

#include <png.h>

//*****************************************************************************
// Save texture as PNG
// From Shazz/71M - thanks guys!
//*****************************************************************************
void PngSaveImage( const char* filename, const void * data, ETextureFormat pixelformat, u32 pitch, u32 width, u32 height )
{
	FILE* fp = fopen(filename, "wb");
	if (!fp)
	{
		return;
	}

	png_structp png_ptr( png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL) );
	if (!png_ptr)
	{
		fclose(fp);
		return;
	}
	png_infop info_ptr( png_create_info_struct(png_ptr) );
	if (!info_ptr)
	{
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
		fclose(fp);
		return;
	}

	png_init_io(png_ptr, fp);
	png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	png_write_info(png_ptr, info_ptr);

	u8* line = (u8*) malloc(width * 3);

	const u32 *		p32( reinterpret_cast< const u32 * >( data ) );
	const u16 *		p16( reinterpret_cast< const u16 * >( data ) );

	for ( u32 y = 0; y < height; y++ )
	{
		for (u32 i = 0, x = 0; x < width; x++)
		{
			u32 color = 0;
			u8 r = 0, g = 0, b = 0;
			switch (pixelformat)
			{
				case TexFmt_5650:
					color = p16[x];
					r = (color & 0x1f) << 3;
					g = ((color >> 5) & 0x3f) << 2;
					b = ((color >> 11) & 0x1f) << 3;
					break;
				case TexFmt_5551:
					color = p16[x];
					r = (color & 0x1f) << 3;
					g = ((color >> 5) & 0x1f) << 3;
					b = ((color >> 10) & 0x1f) << 3;
					break;
				case TexFmt_4444:
					color = p16[x];
					r = (color & 0xf) << 4;
					g = ((color >> 4) & 0xf) << 4;
					b = ((color >> 8) & 0xf) << 4;
					break;
				case TexFmt_8888:
					color = p32[x];
					r = color & 0xff;
					g = (color >> 8) & 0xff;
					b = (color >> 16) & 0xff;
					break;
			}
			line[i++] = r;
			line[i++] = g;
			line[i++] = b;
		}


		p16 = reinterpret_cast< const u16 * >( reinterpret_cast< const u8 * >( p16 ) + pitch );
		p32 = reinterpret_cast< const u32 * >( reinterpret_cast< const u8 * >( p32 ) + pitch );
		png_write_row(png_ptr, line);
	}
	
	free(line);
	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	fclose(fp);
}
