// Taken from http://svn.ps2dev.org/filedetails.php?repname=psp&path=%2Ftrunk%2Fpspgl%2Fpspgl_vidmem.c&rev=0&sc=0

#include "stdafx.h"
#include "VideoMemoryManager.h"

#include "DaedMathUtil.h"

#include <stdlib.h>
#include <string.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <pspge.h>

struct vidmem_chunk
{
	void *ptr;
	u32 len;
};

namespace
{
	static const u32 UNCACHED_MEMORY_ADDRESS = 0x40000000;
}

//*****************************************************************************
//
//*****************************************************************************
class IVideoMemoryManager : public CVideoMemoryManager
{
public:
	IVideoMemoryManager();
	~IVideoMemoryManager();

	void *			InsertNew(u32 idx, u32 adr, u32 size);
	virtual void *	Alloc(u32 size);
	virtual void	Free(void * ptr);

private:
	static	void	DisplayOutOfVideoMemWarning( u32 allocation_size );

private:
	vidmem_chunk *	mpVidMemMap;
	u32				mVidMemMapLen;
};


//*****************************************************************************
//
//*****************************************************************************
template<> bool CSingleton< CVideoMemoryManager >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new IVideoMemoryManager();
	return mpInstance != NULL;
}

//*****************************************************************************
//
//*****************************************************************************
IVideoMemoryManager::IVideoMemoryManager()
:	mpVidMemMap( NULL )
,	mVidMemMapLen( 0 )
{
	pspDebugScreenPrintf( "vram base: %p\n", sceGeEdramGetAddr() );
	pspDebugScreenPrintf( "vram size: %d bytes\n", sceGeEdramGetSize() );
}

//*****************************************************************************
//
//*****************************************************************************
IVideoMemoryManager::~IVideoMemoryManager()
{
}

//*****************************************************************************
//
//*****************************************************************************
void * IVideoMemoryManager::InsertNew(u32 idx, u32 adr, u32 size)
{
	vidmem_chunk *tmp = (vidmem_chunk*)realloc(mpVidMemMap, (mVidMemMapLen + 1) * sizeof(mpVidMemMap[0]));

	if (!tmp)
		return NULL;

	//psp_log("alloc vidmem %lu: adr 0x%08x - size 0x%08x\n", idx, (unsigned int) adr, (unsigned int) size);

	mpVidMemMap = tmp;
	memmove(&mpVidMemMap[idx+1], &mpVidMemMap[idx], (mVidMemMapLen-idx) * sizeof(mpVidMemMap[0]));
	mVidMemMapLen++;
	mpVidMemMap[idx].ptr = (void*) adr;
	mpVidMemMap[idx].len = size;

	return mpVidMemMap[idx].ptr;
}

//*****************************************************************************
//
//*****************************************************************************
void* IVideoMemoryManager::Alloc(u32 size)
{
	// Ensure that all memory is 16-byte aligned
	size = daedalus::AlignPow2( size, 16 );

	u32 start = u32( sceGeEdramGetAddr() ) | UNCACHED_MEMORY_ADDRESS;
	u32 adr = start;
	u32 i;

	for (i=0; i<mVidMemMapLen; i++)
	{
		if (mpVidMemMap[i].ptr != NULL)
		{
			u32 new_adr = (u32) mpVidMemMap[i].ptr;
			if (size <= new_adr - adr)
				return InsertNew(i, adr, size);
			adr = new_adr + mpVidMemMap[i].len;
		}
	}

	if (adr + size > start + sceGeEdramGetSize())
	{
		DisplayOutOfVideoMemWarning( size );
		return NULL;
	}

	return InsertNew(mVidMemMapLen, adr, size);
}

//*****************************************************************************
//
//*****************************************************************************
void  IVideoMemoryManager::Free(void * ptr)
{
	for (u32 i=0; i<mVidMemMapLen; i++)
	{
		if (mpVidMemMap[i].ptr == ptr)
		{
			vidmem_chunk *tmp;

			//psp_log("free vidmem %d: adr 0x%08x - size 0x%08x\n", i, (unsigned int) ptr, (unsigned int) mpVidMemMap[i].len);

			mVidMemMapLen--;
			memmove(&mpVidMemMap[i], &mpVidMemMap[i+1], (mVidMemMapLen-i) * sizeof(mpVidMemMap[0]));
			tmp = (vidmem_chunk*)realloc(mpVidMemMap, mVidMemMapLen * sizeof(mpVidMemMap[0]));
			if (tmp)
				mpVidMemMap = tmp;
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void IVideoMemoryManager::DisplayOutOfVideoMemWarning( u32 allocation_size )
{
#ifndef DAEDALUS_RELEASE_BUILD
	char	msg[ 256 ];
	sprintf( msg, "Out of video memory (allocating %d bytes)", allocation_size );
	DAEDALUS_ASSERTMSG( msg );
#endif
}
